"""
Run Annalist server tests.
"""

from __future__ import print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import sys
import logging
import subprocess

log = logging.getLogger(__name__)

from utils.SetcwdContext    import ChangeCurrentDir

import am_errors
from am_settings            import am_get_settings

def am_runtests(annroot, options):
    """
    Run Annalist server tests.

    annroot     is the root directory for the annalist software installation.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    options.configuration = "runtests"
    testsettings = am_get_settings(annroot, "/nouser/", options)
    if not testsettings:
        print("Settings not found (%s)"%("runtests"), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    log.debug("annalist_root: %s"%(annroot))
    with ChangeCurrentDir(annroot):
        # For some reason, tests are not discovered unless run from here
        cmd = "test"
        subprocess_command = (
            "django-admin %s --pythonpath=%s --settings=%s --top-level-directory=%s"%
            (cmd, annroot, testsettings.modulename, annroot)
            )
        log.debug("am_initialize subprocess: %s"%subprocess_command)
        # OLD: status = os.system(subprocess_command)
        status = subprocess.call(subprocess_command.split())
        log.debug("am_initialize subprocess status: %s"%status)
    return status

# End.



"""
Tests for field rendering functions
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import unittest
import re
from collections import OrderedDict

import logging
log = logging.getLogger(__name__)

from django.conf                        import settings
from django.template                    import Context, Template, loader

from utils.SuppressLoggingContext       import SuppressLogging

from annalist.identifiers               import RDF, RDFS, ANNAL
from annalist.models.site               import Site
from annalist.models.collection         import Collection

from annalist.views.fielddescription    import FieldDescription, field_description_from_view_field

from annalist.views.fields.render_placement     import get_placement_classes
from annalist.views.fields                      import render_repeatgroup
from annalist.views.fields.render_repeatgroup   import RenderRepeatGroup
from annalist.views.fields.render_text          import RenderText
from annalist.views.fields.render_fieldvalue    import RenderFieldValue, get_template
from annalist.views.form_utils.fieldchoice      import FieldChoice

from tests                      import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests                 import init_annalist_test_site, init_annalist_test_coll, resetSitedata
from field_rendering_support    import FieldRendererTestSupport
from entity_testutils       import (
    render_select_options, render_choice_options
    )

#   -----------------------------------------------------------------------------
#
#   FieldRendering tests
#
#   -----------------------------------------------------------------------------

class FieldRenderingTest(FieldRendererTestSupport):
    """
    Tests for Site object interface
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection(self.testsite, "testcoll")
        # self.tokenset_context  = self._make_test_context(
        #     ["aa", "bb", "cc"], repeat_prefix="tokprefix_"
        #     )
        self.intvalue_context  = self._make_test_context(42, repeat_prefix="intprefix_")
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    def _get_repeatgroup_context(self):
        # context['field'] is essentially a bound_field value, combining the repeat group
        # description with a list of entity values to be formatted.
        repeatgroup_context = Context(
            { 'field':
                # ----- Repeated field values, as presented by bound_field
                { 'field_value':
                  [ { "annal:field_id":         "Entity_id"
                    , "annal:field_placement":  "small:0,12;medium:0,6"
                    }
                  , { "annal:field_id":         "Entity_type"
                    , "annal:field_placement":  "small:0,12;medium:6,6"
                    }
                  , { "annal:field_id":         "Entity_label"
                    , "annal:field_placement":  "small:0,12"
                    , "annal:property_uri":     "rdfs:label"
                    }
                  , { "annal:field_id":         "Entity_comment"
                    , "annal:field_placement":  "small:0,12"
                    , "annal:property_uri":     "rdfs:comment"
                    }
                  ]
                # ----- Field description -----
                , 'field_id':                   'View_fields'
                , 'field_name':                 'View_fields'
                , 'field_target_type':          'annal:Field_group'
                , 'field_label':                'Fields'
                , 'field_help':                 'This resource descibes the repeated field description used when displaying and/or editing a record view description'
                , 'field_render_type':          'RepeatGroup'
                , 'field_value_mode':           'Value_direct'
                , 'field_property_uri':         'annal:view_fields'
                , 'field_placement':            get_placement_classes("small:0,12")
                , 'field_placeholder':          '(repeat field description)'
                , 'field_group_ref':            'View_field_view'
                , 'group_id':                   'View_fields'
                , 'group_label':                'Fields'
                , 'group_add_label':            'Add field'
                , 'group_delete_label':         'Remove selected field(s)'
                , 'group_field_descs':
                    [ field_description_from_view_field(
                        self.testcoll, { ANNAL.CURIE.field_id: "Group_field_sel" },        {}
                        )
                    , field_description_from_view_field(
                        self.testcoll, { ANNAL.CURIE.field_id: "Group_field_property" },   {}
                        )
                    , field_description_from_view_field(
                        self.testcoll, { ANNAL.CURIE.field_id: "Group_field_placement" },  {}
                        )
                    ]
                , 'context_extra_values':       {}
                }
            # ----- other values -----
            , 'auth_config':                  True
            })
        return repeatgroup_context

    # Tests

    def test_RenderRepeatGroupTest(self):
        # print "\n".join(sys.path)
        # template = loader.get_template('field/annalist_edit_select.html')
        self.assertEqual(
            RenderRepeatGroup.__name__, "RenderRepeatGroup", 
            "Check RenderRepeatGroup class name"
            )
        return

    def test_RenderRepeatGroupEdit(self):
        fieldrender   = RenderRepeatGroup(render_repeatgroup.edit_group)
        rendered_text = fieldrender.render(self._get_repeatgroup_context())
        # print "\n**************\n"
        # print rendered_text
        # print "\n**************\n"
        # replace runs of whitespace/newlines with single space:
        rendered_text = re.sub(r'\s+', " ", rendered_text)
        field_choices = (
                  [ FieldChoice(id="")
                  , FieldChoice(id="_field/Entity_comment",         label="Comment")
                  , FieldChoice(id="_field/Entity_id",              label="Id")
                  , FieldChoice(id="_field/Entity_label",           label="Label")
                  , FieldChoice(id="_field/Entity_see_also_repeat", label="See also")
                  , FieldChoice(id="_field/Entity_type",            label="Type")
                  ])
        expect_elements = (
            [ '''<div class="group-label small-2 columns"> <span>Fields</span> </div>'''
            , '''<div class="row selectable">'''
            , '''<div class="view-label small-12 medium-4 columns"> <span>Field id</span> </div>'''
            , '''<div class="view-label small-12 medium-4 columns"> <span>Property</span> </div>'''
            , '''<div class="view-label small-12 medium-4 columns"> <span>Position/size</span> </div>'''
            # 1st field
            , '''<input type="checkbox" name="View_fields__select_fields"'''+
              ''' value="0" class="right" />'''
            , re.sub(r'\s+', " ", 
                render_choice_options("View_fields__0__Field_id", field_choices, 
                  "_field/Entity_id", placeholder="(field sel)"
                ))
            , '''<input type="text" size="64" name="View_fields__0__Field_property"'''+
              ''' placeholder="(field URI or CURIE)"'''+
              ''' value=""/>'''
            # 2nd field
            , '''<input type="checkbox" name="View_fields__select_fields"'''+
              ''' value="1" class="right" />'''
            , re.sub(r'\s+', " ", 
                render_choice_options("View_fields__1__Field_id", field_choices, 
                  "_field/Entity_type", placeholder="(field sel)"
                ))
            , '''<input type="text" size="64" name="View_fields__1__Field_property"'''+
              ''' placeholder="(field URI or CURIE)"'''+
              ''' value=""/>'''
            # 3rd field
            , '''<input type="checkbox" name="View_fields__select_fields"'''+
              ''' value="2" class="right" />'''
            , re.sub(r'\s+', " ", 
                render_choice_options("View_fields__2__Field_id", field_choices, 
                  "_field/Entity_label", placeholder="(field sel)"
                ))
            , '''<input type="text" size="64" name="View_fields__2__Field_property"'''+
              ''' placeholder="(field URI or CURIE)"'''+
              ''' value="rdfs:label"/>'''
            # 4th field
            , '''<input type="checkbox" name="View_fields__select_fields"'''+
              ''' value="3" class="right" />'''
            , re.sub(r'\s+', " ", 
                render_choice_options("View_fields__3__Field_id", field_choices, 
                  "_field/Entity_comment", placeholder="(field sel)"
                ))
            , '''<input type="text" size="64" name="View_fields__3__Field_property"'''+
              ''' placeholder="(field URI or CURIE)"'''+
              ''' value="rdfs:comment"/>'''
            # Buttons
            , '''<input type="submit" name="View_fields__remove" value="Remove selected field(s)" />'''
            , '''<input type="submit" name="View_fields__add" value="Add field" />'''
            ])
        for e in expect_elements:
            self.assertInIgnoreWS(e, rendered_text)
        return

    def test_RenderFieldValue_renderers(self):

        class render_int_view(object):
            def render(self, context):
                return str(context['field']['field_value'])

        class render_int_edit(object):
            def render(self, context):
                return (
                    ('''<input type="text" size="64" '''+
                        '''name="%(repeat_prefix)s%(field_name)s" '''+
                        '''placeholder="%(field_placeholder)s" '''+
                        '''value="%(field_value)d"/>''')%
                    { 'field_value':        context['field']['field_value']
                    , 'field_name':         context['field']['field_name']
                    , 'field_placeholder':  context['field']['field_placeholder']
                    , 'repeat_prefix':      context['repeat_prefix']
                    })

        fieldrender = RenderFieldValue(
            view_renderer=render_int_view(), 
            edit_renderer=render_int_edit()
            )
        self._check_value_renderer_results(
            fieldrender,
            context=self.intvalue_context,
            expect_rendered_view="42",
            expect_rendered_edit=
                '''<input type="text" size="64" name="intprefix_test_field" '''+
                '''placeholder="(test placeholder)" value="42"/>'''           
            )
        return

    # Rendering using compiled template supplied
    def test_RenderFieldValue_templates(self):
        view_template = get_template(
            "field/annalist_view_text.html",  
            "Can't load view template"
            )
        edit_template = get_template(
            "field/annalist_edit_text.html",
            "Can't load edit template"
            )
        fieldrender = RenderFieldValue(
            view_template=view_template, 
            edit_template=edit_template
            )
        self._check_value_renderer_results(
            fieldrender,
            context=self.intvalue_context,
            expect_rendered_view=
                '''<span>42</span>''',
            expect_rendered_edit=
                '''<input type="text" size="64" name="intprefix_test_field" \n'''+
                '''       placeholder="(test placeholder)"\n'''+
                '''       value="42" />'''
            )
        return

    # Rendering using template file name supplied
    def test_RenderFieldValue_files(self):
        fieldrender = RenderFieldValue(
            view_file="field/annalist_view_text.html", 
            edit_file="field/annalist_edit_text.html"
            )
        self._check_value_renderer_results(
            fieldrender,
            context=self.intvalue_context,
            expect_rendered_view=
                '''<span>42</span>''',
            expect_rendered_edit=
                '''<input type="text" size="64" name="intprefix_test_field" \n'''+
                '''       placeholder="(test placeholder)"\n'''+
                '''       value="42" />'''
            )
        return


# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
Record field data functions to support entity data testing
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import urlparse

import logging
log = logging.getLogger(__name__)

from django.conf                import settings
from django.http                import QueryDict
from django.utils.http          import urlquote, urlunquote
from django.core.urlresolvers   import resolve, reverse

from annalist.util              import valid_id
from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist                   import layout

from annalist.views.fields.render_placement import (
    get_placement_classes
    )

from entity_testentitydata      import entitydata_list_type_url
from tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from entity_testutils import (
    collection_dir, 
    collection_edit_url,
    collection_entity_view_url,
    site_title
    )

#   -----------------------------------------------------------------------------
#
#   Directory generating functions
#
#   -----------------------------------------------------------------------------

def recordtype_dir(coll_id="testcoll", type_id="testtype"):
    return collection_dir(coll_id) + layout.COLL_TYPE_PATH%{'id': type_id} + "/"

#   -----------------------------------------------------------------------------
#
#   URI generating functions
#
#   -----------------------------------------------------------------------------

#   These use the Django `reverse` function so they correspond to
#   the declared URI patterns.

def recordtype_site_url(site, type_id="testtype"):
    return site._entityurl + layout.SITE_TYPE_PATH%{'id': type_id} + "/"

def recordtype_coll_url(site, coll_id="testcoll", type_id="testtype"):
    return site._entityurl + layout.SITE_COLL_PATH%{'id': coll_id} + "/" + layout.COLL_TYPE_PATH%{'id': type_id} + "/"

def recordtype_url(coll_id="testcoll", type_id="testtype"):
    """
    URL for record type description data; also view using default entity view
    """
    if not valid_id(type_id):
        return None
    return collection_entity_view_url(coll_id=coll_id, type_id="_type", entity_id=type_id)

def recordtype_edit_url(action=None, coll_id=None, type_id=None):
    """
    URI for record type description editing view
    """
    viewname = ( 
        'AnnalistEntityDataView'        if action == "view"   else
        'AnnalistEntityNewView'         if action == "new"    else
        'AnnalistEntityEditView'        if action == "copy"   else
        'AnnalistEntityEditView'        if action == "edit"   else
        'AnnalistRecordTypeDeleteView'  if action == "delete" else
        'unknown'
        )
    kwargs = {'coll_id': coll_id}
    if action != "delete":
        kwargs.update({'action': action, 'type_id': "_type", 'view_id': "Type_view"})
    if type_id:
        if valid_id(type_id):
            kwargs.update({'entity_id': type_id})
        else:
            kwargs.update({'entity_id': "___"})
    return reverse(viewname, kwargs=kwargs)

#   -----------------------------------------------------------------------------
#
#   ----- RecordType data
#
#   -----------------------------------------------------------------------------

def recordtype_value_keys(type_uri=False, supertype_uris=False):
    ks = set(
        [ 'annal:id', 'annal:type_id'
        , 'annal:type'
        , 'annal:url'
        , 'rdfs:label', 'rdfs:comment'
        , 'annal:type_view', 'annal:type_list'
        ])
    if type_uri:
        ks.add('annal:uri')
    if supertype_uris:
        ks.add('annal:supertype_uris')
    return ks

def recordtype_load_keys(type_uri=False, supertype_uris=False):
    return (
        recordtype_value_keys(type_uri=type_uri, supertype_uris=supertype_uris) | 
        {'@id', '@type', '@context'}
        )

def recordtype_create_values(
        coll_id="testcoll", type_id="testtype", update="RecordType",
        type_uri=None, supertype_uris=None
        ):
    """
    Entity values used when creating a record type entity
    """
    d = (
        { 'annal:type':         "annal:Type"
        , 'rdfs:label':         "%s %s/%s"%(update, coll_id, type_id)
        , 'rdfs:comment':       "%s help for %s in collection %s"%(update, type_id, coll_id)
        , 'annal:type_view':    "_view/Default_view"
        , 'annal:type_list':    "_list/Default_list"
        })
    if type_uri:
        d['annal:uri'] = type_uri
        if supertype_uris is not None:
            d['annal:supertype_uris'] = (
                [ { 'annal:supertype_uri': st } for st in supertype_uris ]
                )
        else:
            d['annal:supertype_uris'] = (
                [ { 'annal:supertype_uri': type_uri+"/super1" }
                , { 'annal:supertype_uri': type_uri+"/super2" }
                ])
    return d

def recordtype_values(
        coll_id="testcoll", type_id="testtype", type_uri=None,
        update="RecordType", hosturi=TestHostUri):
    type_url = recordtype_url(coll_id=coll_id, type_id=type_id)
    d = recordtype_create_values(
        coll_id, type_id, update=update, type_uri=type_uri
        ).copy()
    d.update(
        { 'annal:id':       type_id
        , 'annal:type_id':  "_type"
        , 'annal:url':      type_url
        })
    return d

def recordtype_read_values(
        coll_id="testcoll", type_id="testtype", 
        update="RecordType", hosturi=TestHostUri):
    d = recordtype_values(coll_id, type_id,
        update=update,
        hosturi=hosturi
        ).copy()
    d.update(
        { '@id':            "./"
        , '@type':          ["annal:Type"]
        # , '@base':          "../.."
        , '@context':       ["../../coll_context.jsonld"]
        })
    return d

#   -----------------------------------------------------------------------------
#
#   ----- Entity data in recordtype view
#
#   -----------------------------------------------------------------------------

def recordtype_entity_view_context_data(
        coll_id="testcoll", type_id=None, orig_id=None, type_ids=[],
        action=None, update="RecordType",
        type_uri=None, supertype_uris=None #@@ need to deal with these.
    ):
    context_dict = (
        { 'title':              "Collection %s"%(coll_id)
        , 'coll_id':            coll_id
        , 'type_id':            '_type'
        , 'orig_id':            'orig_type_id'
        , 'fields':
          [ { 'field_id':           'Type_id'
            , 'field_name':         'entity_id'
            , 'field_target_type':  'annal:Slug'
            , 'field_label':        'Id'
            , 'field_render_type':  'EntityId'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12;medium:0,6')
            # , 'field_value':      (Supplied separately)
            , 'options':            []
            }
          , { 'field_id':           'Type_label'
            , 'field_name':         'Type_label'
            , 'field_target_type':  'annal:Text'
            , 'field_label':        'Label'
            , 'field_render_type':  'Text'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12')
            , 'field_value':        '%s data ... (%s/%s)'%(update, coll_id, type_id)
            , 'options':            []
            }
          , { 'field_id':           'Type_comment'
            , 'field_name':         'Type_comment'
            , 'field_label':        'Comment'
            , 'field_target_type':  'annal:Richtext'
            , 'field_render_type':  'Markdown'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12')
            , 'field_value':        '%s description ... (%s/%s)'%(update, coll_id, type_id)
            , 'options':            []
            }
          , { 'field_id':           'Type_uri'
            , 'field_name':         'Type_uri'
            , 'field_target_type':  'annal:Identifier'
            , 'field_label':        'URI'
            , 'field_render_type':  'Identifier'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12')
            # , 'field_value':      (Supplied separately)
            , 'options':            []
            }
          , { 'field_id':           'Type_supertype_uris'
            , 'field_name':         'Type_supertype_uris'
            , 'field_target_type':  'annal:Type_supertype_uri'
            , 'field_label':        'Supertype URIs'
            , 'field_render_type':  'RepeatGroupRow'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12')
            # , 'field_value':      (Supplied separately)
            , 'options':            []
            }
          , { 'field_id':           'Type_view'
            , 'field_name':         'Type_view'
            , 'field_target_type':  'annal:View'
            , 'field_label':        'Default view'
            , 'field_render_type':  'Enum_optional'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12;medium:0,6')
            # , 'field_value':      (Supplied separately)
            , 'options':            []
            }
          , { 'field_id':           'Type_list'
            , 'field_name':         'Type_list'
            , 'field_target_type':  'annal:List'
            , 'field_label':        'Default list'
            , 'field_render_type':  'Enum_optional'
            , 'field_value_mode':   'Value_direct'
            , 'field_placement':    get_placement_classes('small:0,12;medium:6,6')
            # , 'field_value':      (Supplied separately)
            , 'options':            []
            }
          ]
        , 'continuation_url':   entitydata_list_type_url(coll_id, "_type")
        })
    if type_id:
        type_url = recordtype_url(coll_id=coll_id, type_id=type_id)
        context_dict['fields'][0]['field_value'] = type_id
        context_dict['fields'][1]['field_value'] = '%s %s/%s'%(update, coll_id, type_id)
        context_dict['fields'][2]['field_value'] = '%s help for %s in collection %s'%(update, type_id, coll_id)
        context_dict['fields'][3]['field_value'] = type_url or ""
        context_dict['orig_id']     = type_id
    if orig_id:
        context_dict['orig_id']     = orig_id
    if action:  
        context_dict['action']      = action
    return context_dict

def recordtype_entity_view_form_data(
        coll_id="testcoll", 
        type_id=None, orig_id=None, 
        action=None, cancel=None, close=None, edit=None, copy=None, task=None,
        update="RecordType",
        type_uri=None
        ):
    """
    Returns a request dictionary that can be used with the Django test client.
    Per Django documentation, multiple values for a key are provided as a list.
    See: https://docs.djangoproject.com/en/1.8/topics/testing/tools/#making-requests
    """
    form_data_dict = (
        { 'Type_label':         '%s data ... (%s/%s)'%(update, coll_id, type_id)
        , 'Type_comment':       '%s description ... (%s/%s)'%(update, coll_id, type_id)
        , 'orig_id':            'orig_type_id'
        , 'continuation_url':   entitydata_list_type_url(coll_id, "_type")
        })
    if type_id:
        type_url  = recordtype_url(coll_id=coll_id, type_id=type_id)
        type_help = '%s help for %s in collection %s'%(update, type_id, coll_id)
        form_data_dict['entity_id']           = type_id
        form_data_dict['orig_id']             = type_id
        form_data_dict['Type_label']          = '%s %s/%s'%(update, coll_id, type_id)
        form_data_dict['Type_comment']        = type_help
        form_data_dict['Type_uri']            = type_url or ""
        form_data_dict['Type_view']           = "_view/Default_view"
        form_data_dict['Type_list']           = "_list/Default_list"
        form_data_dict['orig_type']           = "_type"
    if orig_id:
        form_data_dict['orig_id']   = orig_id
    if action:
        form_data_dict['action']    = action
    if cancel:
        form_data_dict['cancel']    = "Cancel"
    elif close:
        form_data_dict['close']     = "Close"
    elif edit:
        form_data_dict['edit']      = "Edit"
    elif copy:
        form_data_dict['copy']      = "Copy"
    elif task:
        form_data_dict[task]        = task
    else:
        form_data_dict['save']      = "Save"
    if type_uri:
        form_data_dict['Type_uri']                                   = type_uri
        form_data_dict['Type_supertype_uris__0__Type_supertype_uri'] = type_uri+"/super1"
        form_data_dict['Type_supertype_uris__1__Type_supertype_uri'] = type_uri+"/super2"
    return form_data_dict

#   -----------------------------------------------------------------------------
#
#   ----- Recordtype delete confirmation form data
#
#   -----------------------------------------------------------------------------

def recordtype_delete_form_data(type_id=None, list_id="Default_list"):
    return (
        { 'list_choice':        "_list/"+list_id
        , 'continuation_url':   ""
        , 'search_for':         ""
        , 'entity_select':      ["_type/%s"%(type_id)]
        , 'delete':             "Delete"
        })

def recordtype_delete_confirm_form_data(type_id=None):
    return (
        { 'typelist':    type_id,
          'type_delete': 'Delete'
        })

# End.



#!/usr/bin/env python

import os
import sys

if __name__ == "__main__":

    # Select default settings module
    #
    # To specify an alternative:
    #   python manage.py runserver --settings=annalist_site.settings.personal
    # or
    #   export DJANGO_SETTINGS_MODULE=annalist_site.settings.personal

    if 'test' in sys.argv:
        default_settings = "annalist_site.settings.runtests"
    else:
        default_settings = "annalist_site.settings.devel"

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", default_settings)

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)

# End.



"""
Tests for FieldDescription module
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest
from collections import OrderedDict

import logging
log = logging.getLogger(__name__)

from django.conf                        import settings
from django.test                        import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions

from annalist.identifiers               import RDF, RDFS, ANNAL
from annalist.models.site               import Site
from annalist.models.collection         import Collection

from annalist.views.fielddescription    import FieldDescription, field_description_from_view_field

from annalist.views.fields.render_placement     import Placement
from annalist.views.fields.render_repeatgroup   import RenderRepeatGroup
from annalist.views.form_utils.fieldchoice      import FieldChoice

from AnnalistTestCase       import AnnalistTestCase
from tests                  import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests             import init_annalist_test_site, init_annalist_test_coll, resetSitedata
from entity_testentitydata  import entity_url
from entity_testsitedata    import (
    make_field_choices, no_selection,
    get_site_types, get_site_types_sorted, get_site_types_linked,
    get_site_lists, get_site_lists_sorted, get_site_lists_linked,
    get_site_views, get_site_views_sorted, get_site_views_linked,
    get_site_list_types, get_site_list_types_sorted,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    get_site_default_entity_fields, get_site_default_entity_fields_sorted, get_site_default_entity_fields_linked, 
    )

#   -----------------------------------------------------------------------------
#
#   FieldDescription tests
#
#   -----------------------------------------------------------------------------

class FieldDescriptionTest(AnnalistTestCase):
    """
    Tests for Site object interface
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection(self.testsite, "testcoll")
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    def test_FieldDescriptionTest(self):
        self.assertEqual(
            FieldDescription.__name__, "FieldDescription", 
            "Check FieldDescription class name"
            )
        return

    def test_Field_InitialValues(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "_initial_values" }, 
            {}
            )
        expect_placement = Placement(
            field = 'small-12 columns', 
            label = 'small-12 medium-2 columns', 
            value = 'small-12 medium-10 columns'
            )
        expect_field_desc = (
            { 'field_id':                   '_initial_values'
            , 'field_name':                 '_initial_values'
            , 'field_target_type':          ANNAL.CURIE.Text
            , 'field_label':                ''
            , 'field_help':                 ''
            , 'field_render_type':          'Text'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ''
            , 'field_placeholder':          ''
            , 'field_default_value':        None
            , 'field_placement':            expect_placement
            , 'field_ref_type':             None
            , 'field_choices':              None
            , 'field_ref_restriction':      'ALL'
            , 'field_group_ref':            None
            })
        # print repr(fd)
        self.assertDictionaryMatch(fd, expect_field_desc)
        return

    def test_Field_id(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "Field_id" }, 
            {}
            )
        expect_placement = Placement(
            field = 'small-12 medium-6 columns', 
            label = 'small-12 medium-4 columns', 
            value = 'small-12 medium-8 columns'
            )
        expect_field_desc = (
            { 'field_id':                   'Field_id'
            , 'field_name':                 'entity_id'
            , 'field_target_type':          ANNAL.CURIE.Slug
            , 'field_label':                'Id'
            , 'field_render_type':          'EntityId'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ANNAL.CURIE.id
            , 'field_placeholder':          '(field id)'
            , 'field_default_value':        None
            , 'field_placement':            expect_placement
            , 'field_ref_type':             None
            , 'field_choices':              None
            , 'field_ref_restriction':      'ALL'
            , 'field_group_ref':            None
            })
        # print repr(fd)
        self.assertDictionaryMatch(fd, expect_field_desc)
        return

    def test_Group_field_sel(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "Group_field_sel" }, 
            {}
            )
        expect_placement = Placement(
            field='small-12 medium-6 columns', 
            label='small-12 medium-4 columns', 
            value='small-12 medium-8 columns'
            )
        # expect_choice_id_labels = (
        #     [ ('Entity_comment',  "Comment")
        #     , ('Entity_id',       "Id"     )
        #     , ('Entity_label',    "Label"  )
        #     , ('Entity_type',     "Type"   )
        #     ])
        expect_choices = OrderedDict(
            [ (fc.id, fc) 
              for fc in no_selection("(field sel)") + 
                        get_site_default_entity_fields_linked("testcoll") 
            ])
        expect_field_desc = (
            { 'field_id':                   'Group_field_sel'
            , 'field_name':                 'Field_id'
            , 'field_target_type':           ANNAL.CURIE.Slug
            , 'field_label':                'Field id'
            , 'field_render_type':          'Enum_optional'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ANNAL.CURIE.field_id
            , 'field_placeholder':          '(field sel)'
            , 'field_default_value':        ''
            , 'field_placement':            expect_placement
            , 'field_ref_type':             '_field'
            , 'field_choices':              expect_choices
            , 'field_ref_restriction':      'entity[annal:record_type] subtype [annal:field_entity_type]'
            , 'field_group_ref':            None
            })
        # print repr(fd)
        self.assertDictionaryMatch(fd, expect_field_desc)
        return

    def test_View_fields(self):
        fd = field_description_from_view_field(
            self.testcoll, 
            { ANNAL.CURIE.field_id: "View_fields" }, 
            {}
            )
        expect_placement = Placement(
            field = 'small-12 columns', 
            label = 'small-12 medium-2 columns', 
            value = 'small-12 medium-10 columns'
            )
        expect_field_desc = (
            { 'field_id':                   'View_fields'
            , 'field_name':                 'View_fields'
            , 'field_target_type':          ANNAL.CURIE.Field_group
            , 'field_label':                'Fields'
            , 'field_render_type':          'RepeatGroupRow'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ANNAL.CURIE.view_fields
            , 'field_placeholder':          '(repeat field description)'
            , 'field_default_value':        None
            , 'field_placement':            expect_placement
            , 'field_ref_type':             None
            , 'field_choices':              None
            , 'field_ref_restriction':      'ALL'
            , 'field_group_ref':            'View_field_group'
            })
        self.assertDictionaryMatch(fd, expect_field_desc)
        self.assertEqual(fd['field_render_type'], "RepeatGroupRow")
        self.assertEqual(fd['field_render_colhead'], "field/annalist_head_any.html")
        self.assertEqual(fd['field_render_colview'], "field/annalist_item_none.html")
        self.assertIsInstance(fd['field_render_view'], RenderRepeatGroup)
        self.assertIsInstance(fd['field_render_edit'], RenderRepeatGroup)
        expect_group_details = (
            { 'group_id':           "View_fields"
            , 'group_label':        "Fields"
            , 'group_add_label':    "Add field"
            , 'group_delete_label': "Remove selected field(s)"
            })
        self.assertDictionaryMatch(fd, expect_group_details)
        self.assertEqual(len(fd['group_field_descs']), 3)
        # Field type selector
        expect_field0_placement = Placement(
            field='small-12 medium-4 columns', 
            label='small-12 medium-6 columns', 
            value='small-12 medium-6 columns'
            )
        expect_field0_desc = (
            { 'field_id':                   'View_field_sel'
            , 'field_name':                 'Field_id'
            , 'field_target_type':           ANNAL.CURIE.Slug
            , 'field_label':                'Field id'
            , 'field_render_type':          'Enum_optional'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ANNAL.CURIE.field_id
            , 'field_placement':            expect_field0_placement
            })
        self.assertDictionaryMatch(fd['group_field_descs'][0], expect_field0_desc)
        # Field property URI
        expect_field1_placement = Placement(
            field='small-12 medium-4 columns', 
            label='small-12 medium-6 columns', 
            value='small-12 medium-6 columns'
            )
        expect_field1_desc = (
            { 'field_id':                   'View_field_property'
            , 'field_name':                 'Field_property'
            , 'field_target_type':          ANNAL.CURIE.Identifier
            , 'field_label':                'Property'
            , 'field_render_type':          'Identifier'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ANNAL.CURIE.property_uri
            , 'field_placement':            expect_field1_placement
            })
        self.assertDictionaryMatch(fd['group_field_descs'][1], expect_field1_desc)
        # Field placement (within group)
        expect_field2_placement = Placement(
            field='small-12 medium-4 columns', 
            label='small-12 medium-6 columns', 
            value='small-12 medium-6 columns'
            )
        expect_field2_desc = (
            { 'field_id':                   'View_field_placement'
            , 'field_name':                 'Field_placement'
            , 'field_target_type':          ANNAL.CURIE.Placement
            , 'field_label':                'Position/size'
            , 'field_render_type':          'Placement'
            , 'field_value_mode':           'Value_direct'
            , 'field_property_uri':         ANNAL.CURIE.field_placement
            , 'field_placement':            expect_field2_placement
            })
        self.assertDictionaryMatch(fd['group_field_descs'][2], expect_field2_desc)
        return

# End.

if __name__ == "__main__":
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
Create Annalist/Django site data.

Note: uses data in `sampledata/empty/annalist_site`
"""

from __future__ import print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import sys
import logging
import subprocess
import importlib
import shutil

log = logging.getLogger(__name__)

from utils.SetcwdContext            import ChangeCurrentDir
from utils.SuppressLoggingContext   import SuppressLogging

from annalist.identifiers           import ANNAL, RDFS
from annalist.layout                import Layout
from annalist.util                  import removetree, replacetree, updatetree

from annalist.models.site           import Site

import am_errors
from am_settings                    import am_get_settings

def am_createsite(annroot, userhome, options):
    """
    Create Annalist empty site data.

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    options.configuration = "runtests"
    testsettings = am_get_settings(annroot, "/nouser/", options)
    if not testsettings:
        print("Settings not found (%s)"%("runtests"), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status       = am_errors.AM_SUCCESS
    emptysitedir = os.path.join(annroot, "sampledata/empty/annalist_site")
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    sitebasedir  = os.path.join(sitesettings.BASE_DATA_DIR, "annalist_site")
    # Test if old site exists
    if os.path.exists(sitebasedir):
        if options.force:
            # --- Remove old site data from target area
            print("Removing old Annalist site at %s"%(sitebasedir))
            log.info("rmtree: %s"%(sitebasedir))
            removetree(sitebasedir)
        else:
            print("Old data already exists at %s (use --force lor -f to overwrite)"%(sitebasedir), file=sys.stderr)
            return am_errors.AM_EXISTS
    # --- Copy empty site data to target area
    print("Initializing Annalist site in %s"%(sitebasedir))
    log.info("copytree: %s to %s"%(emptysitedir, sitebasedir))
    shutil.copytree(emptysitedir, sitebasedir)
    # --- Copy built-in types and views data to target area
    site_layout = Layout(sitesettings.BASE_DATA_DIR)
    sitedatasrc = os.path.join(annroot, "annalist/sitedata")
    sitedatatgt = os.path.join(sitebasedir, site_layout.SITEDATA_DIR)
    print("Copy Annalist site data from %s to %s"%(sitedatasrc, sitedatatgt))
    for sdir in ("types", "lists", "views", "groups", "fields", "vocabs", "users", "enums"):
        s = os.path.join(sitedatasrc, sdir)
        d = os.path.join(sitedatatgt, sdir)
        print("- %s -> %s"%(sdir, d))
        shutil.copytree(s, d)
    return status

def am_updatesite(annroot, userhome, options):
    """
    Update site data, leaving user data alone

    annroot     is the root directory for the Annalist software installation.
    userhome    is the home directory for the host system user issuing the command.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) > 0:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    sitebasedir  = os.path.join(sitesettings.BASE_DATA_DIR, "annalist_site")
    # Test if site exists
    if not os.path.exists(sitebasedir):
        print("Site %s not found"%(sitebasedir), file=sys.stderr)
        return am_errors.AM_NOTEXISTS
    # --- Copy built-in types and views data to target area
    site_layout = Layout(sitesettings.BASE_DATA_DIR)
    sitedatasrc = os.path.join(annroot, "annalist/sitedata")
    sitedatatgt = os.path.join(sitebasedir, site_layout.SITEDATA_DIR)
    print("Copy Annalist site data from %s to %s"%(sitedatasrc, sitedatatgt))
    for sdir in ("types", "lists", "views", "groups", "fields", "enums"):
        s = os.path.join(sitedatasrc, sdir)
        d = os.path.join(sitedatatgt, sdir)
        print("- %s => %s"%(sdir, d))
        replacetree(s, d)
    for sdir in ("users", "vocabs"):
        s = os.path.join(sitedatasrc, sdir)
        d = os.path.join(sitedatatgt, sdir)
        print("- %s +> %s"%(sdir, d))
        updatetree(s, d)
    # --- (Re)generate site context data
    print("- generating %s"%(site_layout.SITEDATA_CONTEXT_PATH))
    testsite = Site("http://nosite.example.com/", site_layout.SITE_PATH)
    testsite.generate_site_jsonld_context()
    return status

# End.



"""
Tests for entity module
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                import settings
from django.contrib.auth.models import User
from django.test                import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client         import Client

from annalist.identifiers       import ANNAL
from annalist.models.entity     import EntityRoot, Entity

from AnnalistTestCase       import AnnalistTestCase
from tests                  import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests             import init_annalist_test_site, init_annalist_test_coll, resetSitedata

#   -----------------------------------------------------------------------------
#
#   EntityRoot tests
#
#   -----------------------------------------------------------------------------

class TestEntityRootType(EntityRoot):

    _entitytype = "test:EntityRootType"
    _entitypath = None
    _entityfile = ".sub/manifest.jsonld"
    _entityref  = "../"
    _contextref = "../../coll_context.jsonld"

class EntityRootTest(TestCase):
    """
    Tests for EntityRoot object interface
    """

    def setUp(self):
        init_annalist_test_site()
        return

    def tearDown(self):
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    def test_EntityRootTest(self):
        self.assertEqual(EntityRootTest.__name__, "EntityRootTest", "Check class name")
        return

    def test_entityroot_init(self):
        e = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertEqual(e._entitytype,     ANNAL.CURIE.EntityRoot)
        self.assertEqual(e._entityfile,     None)
        self.assertEqual(e._entityref,      None)
        self.assertEqual(e._entityid,       None)
        self.assertEqual(e._entityurl,      TestBaseUri+"/")
        self.assertEqual(e._entitydir,      TestBaseDir+"/")
        self.assertEqual(e._values,         None)
        return

    def test_entityroot_id(self):
        e = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertEqual(e.get_id(),        None)
        e.set_id("testId")
        self.assertEqual(e.get_id(),        "testId")
        return

    def test_entityroot_type_id(self):
        e = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertEqual(e.get_type_id(),   None)
        return

    def test_entityroot_values(self):
        test_values1 = (
            { 'type':           'annal:EntityRoot'
            , 'title':          'Name collection coll1'
            # , 'uri':            '/annalist/coll1'
            })
        test_values1_returned = (
            { 'annal:id':       None
            , 'annal:type_id':  None
            , 'annal:type':     'annal:EntityRoot'
            , 'annal:url':      TestBasePath+'/'
            , 'type':           'annal:EntityRoot'
            , 'title':          'Name collection coll1'
            })
        test_values2 = (
            { 'id':             'TestId'
            , 'type':           'annal:EntityRoot'
            , 'rdfs:label':     'Name collection coll2'
            # , 'uri':            '/annalist/coll2'
            })
        test_values2_returned = (
            { 'annal:id':       None
            , 'annal:type_id':  None
            , 'annal:type':     'annal:EntityRoot'
            , 'annal:url':      TestBasePath+'/'
            , 'id':             'TestId'
            , 'type':           'annal:EntityRoot'
            , 'rdfs:label':     'Name collection coll2'
            })
        e = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertEqual(e.get_values(),    None)
        e.set_values(test_values1)
        self.assertEqual(e.get_values(),    test_values1_returned)
        e.set_values(test_values2)
        self.assertEqual(e.get_values(),    test_values2_returned)
        return

    def test_entityroot_indexing(self):
        test_values1 = (
            { 'type':           'annal:EntityRoot'
            , 'title':          'Name collection coll1'
            # , 'uri':            '/annalist/coll1'
            })
        test_values1_returned = (
            { 'annal:id':       None
            , 'annal:type_id':  None
            , 'annal:type':     'annal:EntityRoot'
            , 'type':           'annal:EntityRoot'
            , 'annal:url':      TestBasePath+'/'
            , 'title':          'Name collection coll1'
            })
        test_values2_returned = (
            { 'annal:id':       None
            , 'annal:type_id':  None
            , 'annal:type':     'annal:EntityRoot'
            , 'annal:url':      TestBasePath+'/'
            , 'id':             'testid'
            , 'type':           'annal:EntityRoot'
            , 'title':          'new title'
            })
        e = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertEqual(e.get_values(),    None)
        e.set_values(test_values1)
        self.assertEqual(set(e.get_values().keys()), set(test_values1_returned.keys()))
        self.assertEqual(e.get_values(),    test_values1_returned)
        for k,v in test_values1_returned.items():
            self.assertEqual(e[k], v)
        e['id']    = "testid"
        e['title'] = "new title" 
        # e['uri']   = "/annalist/coll2"
        self.assertEqual(e.get_values(),    test_values2_returned)
        for k,v in test_values2_returned.items():
            self.assertEqual(e[k], v)
        return

    def test_entityroot_subclass(self):
        e = TestEntityRootType(TestBaseUri, TestBaseDir)
        self.assertEqual(e._entitytype,     "test:EntityRootType")
        self.assertEqual(e._entityfile,     ".sub/manifest.jsonld")
        self.assertEqual(e._entityref,      "../")
        self.assertEqual(e._entityid,       None)
        self.assertEqual(e._entityurl,      TestBaseUri+"/")
        self.assertEqual(e._entitydir,      TestBaseDir+"/")
        self.assertEqual(e._values,         None)
        self.assertEqual(e._dir_path(),     (TestBaseDir+"/.sub", TestBaseDir+"/.sub/manifest.jsonld"))
        return

    def test_entityroot_save_load(self):
        test_values = (
            { 'type':   'annal:EntityRoot'
            , 'title':  'Name collection coll1'
            # , 'uri':    '/annalist/coll1'
            })
        test_values_returned = (
            { '@id':            '../'
            , '@type':          ['test:EntityRootType']
            # , '@base':          "../.."
            , '@context':       ["../../coll_context.jsonld"]
            , 'annal:id':       'testId'
            , 'annal:type_id':  None
            , 'annal:type':     'test:EntityRootType'
            , 'annal:url':      TestBasePath+'/'
            , 'title':          'Name collection coll1'
            , 'type':           'annal:EntityRoot'
            })
        e = TestEntityRootType(TestBaseUri, TestBaseDir)
        e.set_id("testId")
        self.assertEqual(e._entitytype,     "test:EntityRootType")
        self.assertEqual(e._entityfile,     ".sub/manifest.jsonld")
        self.assertEqual(e._entityref,      "../")
        self.assertEqual(e._entityid,       "testId")
        self.assertEqual(e._entityurl,      TestBaseUri+"/")
        self.assertEqual(e._entitydir,      TestBaseDir+"/")
        self.assertEqual(e._values,         None)
        e.set_values(test_values)
        e._save()
        e2 = TestEntityRootType(TestBaseUri, TestBaseDir)
        e2.set_id("testId")
        v2 = e2._load_values()
        self.assertEqual(set(v2.keys()), set(test_values_returned.keys()))
        self.assertEqual(v2, test_values_returned)
        return

    def test_entityroot_exists(self):
        test_values = (
            { 'type':   'annal:EntityRoot'
            , 'title':  'Name collection coll1'
            # , 'uri':    '/annalist/coll1'
            })
        e = TestEntityRootType(TestBaseUri, TestBaseDir)
        e.set_id("testId")
        e.set_values(test_values)
        self.assertFalse(e._exists())
        e._save()
        self.assertTrue(e._exists())
        return

    def test_entityroot_enum_fields(self):
        test_values = (
            { "@type": ['test:EntityRootType']
            , "a": "a"
            , "b": "b"
            , "c": 2
            , "d":
              [ { "d1a": "d1a"
                , "d1b": "d1b"
                }
              , { "d2a": "d2a"
                , "d2b": "d2b"
                }
              ]
            , "e": { "ex": "ex", "ey": "ey"}
            })
        e = TestEntityRootType(TestBaseUri, TestBaseDir)
        e.set_id("testId")
        e.set_values(test_values)
        test_fields_returned = (
            [ (['@type'],           ['test:EntityRootType'])
            , (['annal:id'],        'testId')
            , (['annal:url'],       TestBasePath+'/')
            , (['annal:type_id'],   None)
            , (['annal:type'],      'test:EntityRootType')
            , (["a"],               "a")
            , (["b"],               "b")
            , (["c"],               2)
            , (["d",0,"d1a"],       "d1a")
            , (["d",0,"d1b"],       "d1b")
            , (["d",1,"d2a"],       "d2a")
            , (["d",1,"d2b"],       "d2b")
            , (["e"],               { "ex": "ex", "ey": "ey" })
            ])
        actual_fields_returned = list(e.enum_fields())
        for f in test_fields_returned:
            self.assertIn(f, actual_fields_returned)
        self.assertEqual(len(actual_fields_returned), len(test_fields_returned))
        # Also get get_field...
        for p, v in test_fields_returned:
            self.assertEqual(e.get_field(p), v)
        return


#   -----------------------------------------------------------------------------
#
#   Entity tests
#
#   -----------------------------------------------------------------------------

class TestEntityType(Entity):

    _entitytype = "test:EntityType"
    _entitypath = None
    _entityview = "%(id)s/"
    _entityfile = ".sub/manifest.jsonld"
    _entityref  = "../"
    _contextref = "../../coll_context.jsonld"

class TestEntityTypeSub(Entity):

    _entitytype = "test:EntityTypeSub"
    _entitypath = "sub/%(id)s"
    _entityview = "sub/%(id)s/"
    _entityfile = ".sub/manifest.jsonld"
    _entityref  = "../"
    _contextref = "../../coll_context.jsonld"

class EntityTest(AnnalistTestCase):
    """
    Tests for Entity object interface
    """

    def setUp(self):
        init_annalist_test_site()
        return

    def tearDown(self):
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    #   ----------------------------------------
    #
    #   Helpers
    #
    #   ----------------------------------------

    def values_created(self, entity_type='test:EntityType', entity_title='Name entity test'):
        return (
            { 'type':   entity_type
            , 'title':  entity_title
            })

    def values_returned(self, 
            entity_id='testid', entity_type='test:EntityType', 
            entity_title=None, 
            entity_parent_path=""
            ):
        entity_url = TestBasePath+entity_parent_path+'/%s/'%entity_id
        vals = (
            { 'annal:id':       entity_id
            , 'annal:type_id':  None
            , 'annal:type':     entity_type
            , 'annal:url':      entity_url
            , 'type':           entity_type
            })
        if entity_title:
            vals.update(
                { 'title':          entity_title
                })
        return vals

    def values_reloaded(self, 
            entity_id='testid', entity_type='test:EntityType', 
            entity_title=None, 
            entity_parent_path=""
            ):
        vals = self.values_returned(
            entity_id=entity_id, entity_type=entity_type, 
            entity_title=entity_title,
            entity_parent_path=entity_parent_path
            )
        vals.update(
            { '@id':        '../'   # Local testing only: see `TestEntityType`
            , '@type':      [entity_type]
            # , '@base':      "../.."
            , '@context':   ["../../coll_context.jsonld"]
            })
        return vals

    #   ----------------------------------------
    #
    #   Tests
    #
    #   ----------------------------------------

    def test_EntityTest(self):
        self.assertEqual(EntityTest.__name__, "EntityTest", "Check class name")
        return

    def test_entity_init(self):
        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = Entity(r, "testid")
        self.assertEqual(e._entitytype,     ANNAL.CURIE.Entity)
        self.assertEqual(e._entityfile,     None)
        self.assertEqual(e._entityref,      None)
        self.assertEqual(e._entityid,       "testid")
        self.assertEqual(e._entityurl,      TestBaseUri+"/testid/")
        self.assertEqual(e._entitydir,      TestBaseDir+"/testid/")
        self.assertEqual(e._values,         None)
        return

    def test_entity_type_id(self):
        r = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertEqual(r.get_type_id(),   None)
        e1 = Entity(r, "testid1")
        self.assertEqual(e1.get_type_id(),  None)
        e2 = Entity(e1, "testid2")
        self.assertEqual(e2.get_type_id(),  None)
        return

    def test_entity_subclass(self):
        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = TestEntityType(r, "testid")
        self.assertEqual(e._entitytype,     "test:EntityType")
        self.assertEqual(e._entityfile,     ".sub/manifest.jsonld")
        self.assertEqual(e._entityref,      "../")
        self.assertEqual(e._entityid,       "testid")
        self.assertEqual(e._entityurl,      TestBaseUri+"/testid/")
        self.assertEqual(e._entitydir,      TestBaseDir+"/testid/")
        self.assertEqual(e._values,         None)
        self.assertEqual(e._dir_path(),     (TestBaseDir+"/testid/.sub", TestBaseDir+"/testid/.sub/manifest.jsonld"))
        return

    def test_entity_dict(self):
        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = TestEntityType(r, "testid")
        self.assertEqual(e.get("foo",  "bar"),    "bar")
        self.assertEqual(e.get("type", "notype"), "notype")
        e.set_values({'type': 'annal:EntityRoot'})
        self.assertEqual(e.get("foo",  "bar"),    "bar")
        self.assertEqual(e.get("type", "notype"), "annal:EntityRoot")
        expect_keys = set(
            ['annal:id', 'annal:type_id'
            , 'annal:type'
            , 'annal:url'
            , 'type'
            ])
        self.assertEqual(set(e.keys()), expect_keys)
        expect_items = set(
            [ ('annal:id',      'testid')
            , ('annal:type_id',  None)
            , ('annal:type',    'test:EntityType')
            , ('annal:url',     TestBasePath+'/testid/')
            , ('type',          'annal:EntityRoot')
            ])
        self.assertEqual(set(e.items()), expect_items)
        return

    def test_entity_path(self):
        test_values = ({})
        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = TestEntityType(r, "testid")
        e.set_values(test_values)
        e._save()
        p = TestEntityType.path(r, "testid")
        self.assertEqual(p, os.path.join(TestBaseDir+"/testid", ".sub/manifest.jsonld"))
        return

    def test_entity_create_exists(self):
        test_values = self.values_created(entity_type='test:EntityType', entity_title='Name entity test')
        test_values_returned = self.values_returned(
            entity_id='testid', entity_type='test:EntityType', entity_title='Name entity test'
            )
        r = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertFalse(TestEntityType.exists(r, "testid"))
        e = TestEntityType.create(r, "testid", test_values)
        self.assertTrue(TestEntityType.exists(r, "testid"))
        v = e.get_values()
        self.assertEqual(set(v.keys()), set(test_values_returned.keys()))
        self.assertEqual(v, test_values_returned)
        return

    def test_entity_create_load(self):
        test_values = self.values_created(entity_type='test:EntityType', entity_title='Name entity test2')
        test_values_returned = self.values_reloaded(
            entity_id='testid2', entity_type='test:EntityType', entity_title='Name entity test2'
            )
        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = TestEntityType.create(r, "testid2", test_values)
        e2 = TestEntityType.load(r, "testid2")
        v = e2.get_values()
        self.assertEqual(set(v.keys()), set(test_values_returned.keys()))
        self.assertEqual(v, test_values_returned)
        return

    def test_entity_children(self):
        test_values1 = self.values_created(entity_type='test:EntityType', entity_title='Name entity test1')
        test_values1_returned = self.values_reloaded(
            entity_id='testid1', entity_type='test:EntityType', entity_title='Name entity test1',
            entity_parent_path="/testbase"
            )
        test_values2 = self.values_created(entity_type='test:EntityType', entity_title='Name entity test2')
        test_values2_returned = self.values_reloaded(
            entity_id='testid2', entity_type='test:EntityType', entity_title='Name entity test2',
            entity_parent_path="/testbase"
            )
        r  = EntityRoot(TestBaseUri, TestBaseDir)
        b  = TestEntityType.create(r, "testbase", {})
        e1 = TestEntityType.create(b, "testid1", test_values1)
        e2 = TestEntityType.create(b, "testid2", test_values2)
        eids = list(b.child_entity_ids(TestEntityType))
        self.assertEqual(sorted(eids), ["testid1", "testid2"])
        es  = sorted(b.child_entities(TestEntityType), key=(lambda e: e.get_id()))
        es1 = es[0]
        v1  = es1.get_values()
        # log.info(v1)
        # log.info(test_values1_returned)
        self.assertEqual(set(v1.keys()), set(test_values1_returned.keys()))
        self.assertDictionaryMatch(v1, test_values1_returned)
        es2 = es[1]
        v2  = es2.get_values()
        self.assertEqual(set(v2.keys()), set(test_values2_returned.keys()))
        self.assertDictionaryMatch(v2, test_values2_returned)
        return

    def test_entity_create_remove(self):
        test_values = self.values_created(entity_type='test:EntityType', entity_title='Name entity test')
        r = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertFalse(TestEntityType.exists(r, "testid3"))
        e = TestEntityType.create(r, "testid3", test_values)
        self.assertTrue(TestEntityType.exists(r, "testid3"))
        s = TestEntityType.remove(r, "testid3")
        self.assertFalse(TestEntityType.exists(r, "testid3"))
        return

    # The following tests repeat the above using an entity class with 
    # an explcit relative path from parent to entity.

    def test_entity_sub_path(self):
        test_values = self.values_created(entity_type='annal:EntityRoot', entity_title='Name collection coll1')
        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = TestEntityTypeSub(r, "testid")
        self.assertEqual(e._entitydir, TestBaseDir+"/sub/testid/")
        e.set_values(test_values)
        e._save()
        p = TestEntityTypeSub.path(r, "testid")
        self.assertEqual(p, os.path.join(TestBaseDir+"/sub/testid", ".sub/manifest.jsonld"))
        return

    def test_entity_sub_create_exists(self):
        test_values = self.values_created(entity_type='test:EntityTypeSub', entity_title='Name entity test')
        test_values_returned = self.values_returned(
            entity_id='testid', entity_type='test:EntityTypeSub', entity_title='Name entity test',
            entity_parent_path="/sub"
            )
        r = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertFalse(TestEntityTypeSub.exists(r, "testid"))
        e = TestEntityTypeSub.create(r, "testid", test_values)
        self.assertTrue(TestEntityTypeSub.exists(r, "testid"))
        p = TestEntityTypeSub.path(r, "testid")
        self.assertEqual(p, os.path.join(TestBaseDir+"/sub/testid", ".sub/manifest.jsonld"))
        v = e.get_values()
        self.assertEqual(set(v.keys()), set(test_values_returned.keys()))
        self.assertEqual(v, test_values_returned)
        return

    def test_entity_sub_create_load(self):
        test_values = self.values_created(entity_type='test:EntityTypeSub', entity_title='Name entity test2')
        test_values_returned = self.values_reloaded(
            entity_id='testid2', entity_type='test:EntityTypeSub', entity_title='Name entity test2',
            entity_parent_path="/sub"
            )

        r = EntityRoot(TestBaseUri, TestBaseDir)
        e = TestEntityTypeSub.create(r, "testid2", test_values)
        e2 = TestEntityTypeSub.load(r, "testid2")
        v = e2.get_values()
        self.assertEqual(set(v.keys()), set(test_values_returned.keys()))
        self.assertEqual(v, test_values_returned)
        return

    def test_entity_sub_create_remove(self):
        test_values = self.values_created(entity_type='test:EntityTypeSub', entity_title='Name entity test')
        r = EntityRoot(TestBaseUri, TestBaseDir)
        self.assertFalse(TestEntityTypeSub.exists(r, "testid3"))
        e = TestEntityTypeSub.create(r, "testid3", test_values)
        self.assertTrue(TestEntityTypeSub.exists(r, "testid3"))
        s = TestEntityTypeSub.remove(r, "testid3")
        self.assertFalse(TestEntityTypeSub.exists(r, "testid3"))
        return

    def test_entity_allocate_id(self):
        test_values = self.values_created(entity_type='test:EntityType', entity_title='Name entity test')
        r = EntityRoot(TestBaseUri, TestBaseDir)
        eid = TestEntityType.allocate_new_id(r)
        self.assertEqual(eid, "00000001")
        self.assertFalse(TestEntityType.exists(r, eid))
        eid = TestEntityType.allocate_new_id(r)
        self.assertEqual(eid, "00000001")
        e = TestEntityType.create(r, eid, test_values)
        self.assertTrue(TestEntityType.exists(r, eid))
        eid = TestEntityType.allocate_new_id(r)
        self.assertEqual(eid, "00000002")
        return

# End.



"""
Annalist site root URL definitions for testing
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

from django.conf        import settings
from django.conf.urls   import patterns, include, url

urlpatterns = patterns('',
    # Examples:
    # url(r'^$', 'annalist_site.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),
    url(r'^'+settings.TEST_BASE_PATH+"/", include('annalist.urls')),
)



"""
This module contains definitions intended to reflect the site-wide data
used by all Annalist installations.

Test cases should use values returned by this module so that additions to 
the site data can be updated here, in just one place.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import urlparse

import logging
log = logging.getLogger(__name__)

from annalist.util                          import valid_id, split_type_entity_id

from annalist.views.form_utils.fieldchoice  import FieldChoice

from entity_testentitydata                  import entity_url

#   -----------------------------------------------------------------------------
#
#   ----- Field choice generation support functions
#
#   -----------------------------------------------------------------------------

# Convert list of ids and labvels into list of field choices
def make_field_choices(options):
    return [ FieldChoice(v, v, l) for v,l in options ]

def no_selection(label):
    return [ FieldChoice("", "", label) ]

def id_from_field_choice(fc):
    type_id, entity_id = split_type_entity_id(fc.id)
    return entity_id

def add_link_to_field_choice(fc, coll_id, default_type_id=None):
    type_id, entity_id = split_type_entity_id(fc.id, default_type_id=default_type_id)
    return fc.add_link(entity_url(coll_id, type_id, entity_id))

#   -----------------------------------------------------------------------------
#
#   ----- Site-wide entities defined
#
#   -----------------------------------------------------------------------------

#   ----- Types -----

site_types = (
    [ FieldChoice("_type/_initial_values")
    , FieldChoice("_type/_field",                     label="View field"                  )
    , FieldChoice("_type/_group",                     label="Field group"                 )
    , FieldChoice("_type/_list",                      label="List"                        )
    , FieldChoice("_type/_type",                      label="Type"                        )
    , FieldChoice("_type/_user",                      label="User permissions"            )
    , FieldChoice("_type/_view",                      label="View"                        )
    , FieldChoice("_type/_vocab",                     label="Vocab namespace"             )
    , FieldChoice("_type/BibEntry_type",              label="Bibliographic record"        )
    , FieldChoice("_type/Default_type",               label="Default record"              )
    , FieldChoice("_type/Enum_bib_type",              label="Bibliographic entry type"    )
    , FieldChoice("_type/Enum_list_type",             label="List display type"           )
    , FieldChoice("_type/Enum_render_type",           label="Field render type"           )
    , FieldChoice("_type/Enum_value_mode",            label="Field value mode"            )
    ])

def get_site_types_sorted():
    return site_types[1:]

def get_site_types_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "_type") 
          for fc in get_site_types_sorted() 
        ])

def get_site_types():
    return set( ( id_from_field_choice(fc) for fc in get_site_types_sorted() )  )

#   ----- Lists -----

site_lists = (
    [ FieldChoice("_list/_initial_values")
    , FieldChoice("_list/BibEntry_list",              label="List bibliographic entries")
    , FieldChoice("_list/Default_list",               label="List entities")
    , FieldChoice("_list/Default_list_all",           label="List entities with type information")
    , FieldChoice("_list/Field_group_list",           label="List field groups")
    , FieldChoice("_list/Field_list",                 label="List fields")
    , FieldChoice("_list/List_list",                  label="List lists")
    , FieldChoice("_list/Type_list",                  label="List types")
    , FieldChoice("_list/User_list",                  label="User permissions")
    , FieldChoice("_list/View_list",                  label="List views")
    , FieldChoice("_list/Vocab_list",                 label="List vocabulary namespaces")
    ])

def get_site_lists_sorted():
    return site_lists[1:]

def get_site_lists_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "_list")
          for fc in get_site_lists_sorted() 
        ])

def get_site_lists():
    return set( ( id_from_field_choice(fc) for fc in get_site_lists_sorted() )  )

#   ----- List types -----

site_list_types = (
    [ FieldChoice("Enum_list_type/_initial_values")
    , FieldChoice("Enum_list_type/Grid",                       label="Grid display")
    , FieldChoice("Enum_list_type/List",                       label="List display")
    ])

def get_site_list_types_sorted():
    return site_list_types[1:]

def get_site_list_types():
    return set( ( fc.id for fc in get_site_list_types_sorted() )  )

#   ----- Views -----

site_views = (
    [ FieldChoice("_view/_initial_values")
    , FieldChoice("_view/BibEntry_view",              label="Bibliographic metadata"  )
    , FieldChoice("_view/Default_view",               label="Default record view"     )
    , FieldChoice("_view/Field_group_view",           label="Field group view"        )
    , FieldChoice("_view/Field_view",                 label="Field description view"  )
    , FieldChoice("_view/List_view",                  label="List description view"   )
    , FieldChoice("_view/Type_view",                  label="Type description view"   )
    , FieldChoice("_view/User_view",                  label="User permissions view"   )
    , FieldChoice("_view/View_view",                  label="View description view"   )
    , FieldChoice("_view/Vocab_view",                 label="Vocabulary namespace view")
    ])

def get_site_views_sorted():
    return site_views[1:]

def get_site_views_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "_view") 
          for fc in get_site_views_sorted() 
        ])

def get_site_views():
    return set( ( id_from_field_choice(fc) for fc in get_site_views_sorted() )  )

#   ----- Field groups -----

site_field_groups = (
    [ FieldChoice("_group/_initial_values")
    , FieldChoice("_group/Bib_book_group",             label="BibEntry book fields"       )
    , FieldChoice("_group/Bib_identifier_group",       label="BibEntry identifier fields" )
    , FieldChoice("_group/Bib_journal_group",          label="BibEntry journal fields"    )
    , FieldChoice("_group/Bib_license_group",          label="BibEntry license fields"    )
    , FieldChoice("_group/Bib_person_group",           label="BibEntry person fields"     )
    , FieldChoice("_group/Bib_publication_group",      label="BibEntry publication fields")
    , FieldChoice("_group/Entity_see_also_repeat",     label="Links to further information")
    , FieldChoice("_group/Group_field_group",          label="Group field fields"         )
    , FieldChoice("_group/List_field_group",           label="List field fields"          )
    , FieldChoice("_group/Type_alias_group",           label="Field alias fields"         )
    , FieldChoice("_group/Type_supertype_uri_group",   label="Supertype URIs"             )
    , FieldChoice("_group/View_field_group",           label="View field fields"          )
    ]) 

def get_site_field_groups_sorted():
    return site_field_groups[1:]

def get_site_field_groups_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "_group") 
          for fc in get_site_field_groups_sorted()
        ])

def get_site_field_groups():
    return set( ( id_from_field_choice(fc) for fc in get_site_field_groups_sorted() ) )

#   ----- Fields -----

site_default_entity_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Entity_comment",             label="Comment"          )
    , FieldChoice("_field/Entity_id",                  label="Id"               )
    , FieldChoice("_field/Entity_label",               label="Label"            )
    , FieldChoice("_field/Entity_see_also_repeat",     label="See also"         )
    , FieldChoice("_field/Entity_type",                label="Type"             )
    ])

site_bibentry_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Bib_address",                label="Address"             )
    , FieldChoice("_field/Bib_alternate",              label="Alternate name"      )
    , FieldChoice("_field/Bib_authors",                label="Author(s)"           )
    , FieldChoice("_field/Bib_bookentry",              label="Book content"        )
    , FieldChoice("_field/Bib_booktitle",              label="Book title"          )
    , FieldChoice("_field/Bib_chapter",                label="Chapter"             )
    , FieldChoice("_field/Bib_description",            label="Description"         )
    , FieldChoice("_field/Bib_edition",                label="Edition"             )
    , FieldChoice("_field/Bib_editors",                label="Editor(s)"           )
    , FieldChoice("_field/Bib_eprint",                 label="ePrint"              )
    , FieldChoice("_field/Bib_firstname",              label="First name"          )
    , FieldChoice("_field/Bib_howpublished",           label="How published"       )
    , FieldChoice("_field/Bib_id",                     label="Id"                  )
    , FieldChoice("_field/Bib_idanchor",               label="Anchor"              )
    , FieldChoice("_field/Bib_identifiers",            label="Identifiers"         )
    , FieldChoice("_field/Bib_idtype",                 label="Identifier type"     )
    , FieldChoice("_field/Bib_institution",            label="Institution"         )
    , FieldChoice("_field/Bib_journal",                label="Journal"             )
    , FieldChoice("_field/Bib_jurisdiction",           label="Jurisdiction"        )
    , FieldChoice("_field/Bib_lastname",               label="Last name"           )
    , FieldChoice("_field/Bib_license",                label="License"             )
    , FieldChoice("_field/Bib_month",                  label="Month"               )
    , FieldChoice("_field/Bib_name",                   label="Name"                )
    , FieldChoice("_field/Bib_note",                   label="Note"                )
    , FieldChoice("_field/Bib_number",                 label="Issue number"        )
    , FieldChoice("_field/Bib_organization",           label="Organization"        )
    , FieldChoice("_field/Bib_pages",                  label="Pages"               )
    , FieldChoice("_field/Bib_publication_details",    label="Publication details" )
    , FieldChoice("_field/Bib_publisher",              label="Publisher"           )
    , FieldChoice("_field/Bib_school",                 label="School"              )
    , FieldChoice("_field/Bib_shortcode",              label="Short code"          )
    , FieldChoice("_field/Bib_title",                  label="Title"               )
    , FieldChoice("_field/Bib_type",                   label="Publication type"    )
    , FieldChoice("_field/Bib_url",                    label="URL"                 )
    , FieldChoice("_field/Bib_volume",                 label="Volume"              )
    , FieldChoice("_field/Bib_year",                   label="Year"                )
    ])

site_field_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Field_comment",              label="Help"                )
    , FieldChoice("_field/Field_default",              label="Default"             )
    , FieldChoice("_field/Field_entity_type",          label="Entity type"         )
    , FieldChoice("_field/Field_fieldref",             label="Refer to field"      )
    , FieldChoice("_field/Field_groupref",             label="Field group"         )
    , FieldChoice("_field/Field_id",                   label="Id"                  )
    , FieldChoice("_field/Field_label",                label="Label"               )
    , FieldChoice("_field/Field_missing",              label="Missing"             )
    , FieldChoice("_field/Field_placeholder",          label="Placeholder"         )
    , FieldChoice("_field/Field_placement",            label="Position/size"       )
    , FieldChoice("_field/Field_property",             label="Property"            )
    , FieldChoice("_field/Field_render",               label="Field render type"   )
    , FieldChoice("_field/Field_repeat_label_add",     label="Add fields label"    )
    , FieldChoice("_field/Field_repeat_label_delete",  label="Delete fields label" )
    , FieldChoice("_field/Field_restrict",             label="Value restriction"   )
    , FieldChoice("_field/Field_type",                 label="Field value type"    )
    , FieldChoice("_field/Field_typeref",              label="Refer to type"       )
    , FieldChoice("_field/Field_value_mode",           label="Value mode"          )
    ])  

site_group_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Group_comment",              label="Help"                )
    , FieldChoice("_field/Group_fields",               label="Fields"              )
    , FieldChoice("_field/Group_id",                   label="Id"                  )
    , FieldChoice("_field/Group_label",                label="Label"               )
    , FieldChoice("_field/Group_target_type",          label="Record type"         )
    ])

site_group_field_group_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Group_field_placement",      label="Position/size"       )
    , FieldChoice("_field/Group_field_property",       label="Property"            )
    , FieldChoice("_field/Group_field_sel",            label="Field id"            )
    ])

site_list_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/List_comment",               label="Help"                )
    , FieldChoice("_field/List_default_type",          label="Record type"         )
    , FieldChoice("_field/List_default_view",          label="View"                )   
    , FieldChoice("_field/List_entity_selector",       label="Selector"            )
    , FieldChoice("_field/List_fields",                label="Fields"              )
    , FieldChoice("_field/List_id",                    label="Id"                  )
    , FieldChoice("_field/List_label",                 label="Label"               )
    , FieldChoice("_field/List_target_type",           label="Record type URI"     )
    , FieldChoice("_field/List_type",                  label="List display type"   )
    ])
    # , FieldChoice("List_choice",                label="List view"           )

site_list_field_group_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/List_field_placement",       label="Position/size"       )
    , FieldChoice("_field/List_field_property",        label="Property"            )
    , FieldChoice("_field/List_field_sel",             label="Field id"            )
    ])

site_type_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Type_aliases",               label="Field aliases"       )
    , FieldChoice("_field/Type_comment",               label="Comment"             )
    , FieldChoice("_field/Type_id",                    label="Id"                  )
    , FieldChoice("_field/Type_label",                 label="Label"               )
    , FieldChoice("_field/Type_list",                  label="Default list"        )
    , FieldChoice("_field/Type_supertype_uris",        label="Supertype URIs"      )
    , FieldChoice("_field/Type_uri",                   label="URI"                 )
    , FieldChoice("_field/Type_view",                  label="Default view"        )
    ])

site_type_supertype_uris_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Type_supertype_uri",         label="Supertype URI"       )
    ])

site_type_aliases_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Type_alias_source",          label="Type alias source"   )
    , FieldChoice("_field/Type_alias_target",          label="Type alias target"   )
    ])

site_user_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/User_description",           label="Description"         )
    , FieldChoice("_field/User_id",                    label="User Id"             )
    , FieldChoice("_field/User_name",                  label="User name"           )
    , FieldChoice("_field/User_permissions",           label="Permissions"         )
    , FieldChoice("_field/User_uri",                   label="URI"                 )
    ])

site_view_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/View_comment",               label="Help"                )
    , FieldChoice("_field/View_edit_view",             label="Editable view?"      )
    , FieldChoice("_field/View_fields",                label="Fields"              )
    , FieldChoice("_field/View_id",                    label="Id"                  )
    , FieldChoice("_field/View_label",                 label="Label"               )
    , FieldChoice("_field/View_target_type",           label="Record type"         )
    ])
    # , FieldChoice("View_choice",          "Choose view")

site_view_field_group_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/View_field_placement",       label="Position/size"       )
    , FieldChoice("_field/View_field_property",        label="Property"            )
    , FieldChoice("_field/View_field_sel",             label="Field id"            )
    ])

site_vocab_fields = (
    [ FieldChoice("_field/_initial_values")
    , FieldChoice("_field/Vocab_uri",                  label="Vocabulary URI"      )
    ])

site_fields = (
    [ FieldChoice("_field/_initial_values")] +
    site_bibentry_fields[1:] +
    site_default_entity_fields[1:] +
    site_field_fields[1:] +
    site_group_fields[1:] +
    site_list_fields[1:] +
    site_type_fields[1:] +
    site_user_fields[1:] +
    site_view_fields[1:] +
    site_vocab_fields[1:] +
    [])

def get_site_fields_sorted():
    return site_fields[1:]

def get_site_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_fields_sorted() ) )

def get_site_default_entity_fields_sorted():
    return site_default_entity_fields[1:]

def get_site_default_entity_fields_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "_field") 
          for fc in get_site_default_entity_fields_sorted() 
        ])

def get_site_default_entity_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_default_entity_fields_sorted() ) )

def get_site_default_entity_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_default_entity_fields_sorted() ) )

def get_site_view_fields_sorted():
    return site_default_entity_fields[1:] + site_view_fields[1:]

def get_site_view_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_view_fields_sorted() ) )

def get_site_field_fields_sorted():
    return site_default_entity_fields[1:] + site_field_fields[1:]

def get_site_field_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_field_fields_sorted() ) )

def get_site_group_fields_sorted():
    return site_default_entity_fields[1:] + site_group_fields[1:]

def get_site_group_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_group_fields_sorted() ) )

def get_site_list_fields_sorted():
    return site_default_entity_fields[1:] + site_list_fields[1:]

def get_site_list_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_list_fields_sorted() ) )

def get_site_type_fields_sorted():
    return site_default_entity_fields[1:] + site_type_fields[1:]

def get_site_type_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_type_fields_sorted() ) )

def get_site_user_fields_sorted():
    return site_default_entity_fields[1:] + site_user_fields[1:]

def get_site_user_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_user_fields_sorted() ) )

def get_site_vocab_fields_sorted():
    return ( 
        site_default_entity_fields[1:] + 
        site_vocab_fields[1:]
        )

def get_site_vocab_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_vocab_fields_sorted() ) )

def get_site_bibentry_fields_sorted():
    return site_bibentry_fields[1:] + site_default_entity_fields[1:]

def get_site_bibentry_fields():
    return set( ( id_from_field_choice(fc) for fc in get_site_bibentry_fields_sorted() ) )

#   ----- Field render types -----

site_field_types = (
    [ FieldChoice("Enum_render_type/_initial_values")
    , FieldChoice("Enum_render_type/CheckBox",          label="CheckBox"                     )
    , FieldChoice("Enum_render_type/EntityId",          label="Entity Id"                    )
    , FieldChoice("Enum_render_type/EntityTypeId",      label="Entity type Id"               )
    , FieldChoice("Enum_render_type/Enum",              label="Required entity ref"          )
    , FieldChoice("Enum_render_type/Enum_choice",       label="Entity choice"                )
    , FieldChoice("Enum_render_type/Enum_optional",     label="Optional entity ref"          )
    , FieldChoice("Enum_render_type/FileUpload",        label="File upload"                  )
    , FieldChoice("Enum_render_type/Identifier",        label="Identifier"                   )
    , FieldChoice("Enum_render_type/Markdown",          label="Markdown rich text"           )
    , FieldChoice("Enum_render_type/Placement",         label="Position/size"                )
    , FieldChoice("Enum_render_type/RefAudio",          label="Ref audio file"               )
    , FieldChoice("Enum_render_type/RefImage",          label="Ref image file"               )
    , FieldChoice("Enum_render_type/RefMultifield",     label="Fields of referenced entity"  )
    , FieldChoice("Enum_render_type/RepeatGroup",       label="Repeating field group"        )
    , FieldChoice("Enum_render_type/RepeatGroupRow",    label="Repeating fields as row"      )
    , FieldChoice("Enum_render_type/Slug",              label="Short name"                   )
    , FieldChoice("Enum_render_type/Text",              label="Short text"                   )
    , FieldChoice("Enum_render_type/Textarea",          label="Multiline text"               )
    , FieldChoice("Enum_render_type/TokenSet",          label="Space-separated tokens"       )
    , FieldChoice("Enum_render_type/URIImport",         label="Web import"                   )
    , FieldChoice("Enum_render_type/URILink",           label="Web link"                     )
    ])
    # , FieldChoice("Enum_render_type/Field",             label="Field Id"             ) #@@deprecated
    # , FieldChoice("Enum_render_type/List",              label="List Id"              ) #@@deprecated
    # , FieldChoice("Enum_render_type/Type",              label="Type Id"              ) #@@deprecated
    # , FieldChoice("Enum_render_type/View",              label="View Id"              ) #@@deprecated

def get_site_field_types_sorted():
    return site_field_types[1:]

def get_site_field_types_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "Enum_render_type") 
          for fc in get_site_field_types_sorted()
        ])

def get_site_field_types():
    return set( ( id_from_field_choice(fc) for fc in get_site_field_types_sorted() ) )

#   ----- Vocabulary namespaces -----

site_vocabs = (
    [ FieldChoice("_vocab/_initial_values")
    , FieldChoice("_vocab/annal",      label="Vocabulary namespace for Annalist-defined terms")
    , FieldChoice("_vocab/owl",        label="OWL ontology namespace")
    , FieldChoice("_vocab/rdf",        label="RDF core namespace")
    , FieldChoice("_vocab/rdfs",       label="RDF schema namespace")
    , FieldChoice("_vocab/xsd",        label="XML Schema datatypes namespace")
    ])

def get_site_vocabs_sorted():
    return site_vocabs[1:]

def get_site_vocabs_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "_vocab") 
          for fc in get_site_vocabs_sorted()
        ])

def get_site_vocabs():
    return set( ( id_from_field_choice(fc) for fc in get_site_vocabs_sorted() )  )

#   ----- Field value mode types -----

site_value_modes = (
    [ FieldChoice("Enum_value_mode/_initial_values")
    , FieldChoice("Enum_value_mode/Value_direct",               label="Direct value"      )
    , FieldChoice("Enum_value_mode/Value_entity",               label="Entity reference"  )
    , FieldChoice("Enum_value_mode/Value_field",                label="Field reference"   )
    , FieldChoice("Enum_value_mode/Value_import",               label="Import from web"   )
    , FieldChoice("Enum_value_mode/Value_upload",               label="File upload"       )
    ])

def get_site_value_modes_sorted():
    return site_value_modes[1:]

def get_site_value_modes_linked(coll_id):
    return (
        [ add_link_to_field_choice(fc, coll_id, "Enum_value_mode") 
          for fc in get_site_value_modes_sorted()
        ])

def get_site_value_modes():
    return set( ( id_from_field_choice(fc) for fc in get_site_value_modes_sorted() )  )

# End.



"""
Annalist record list

A record type is represented by:
- an ID (slug)
- a URI
- a name/label
- a description
- ...
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

class RecordList(EntityData):

    _entitytype     = ANNAL.CURIE.List
    _entitytypeid   = "_list"
    _entityview     = layout.COLL_LIST_VIEW
    _entitypath     = layout.COLL_LIST_PATH
    _entityaltpath  = layout.SITE_LIST_PATH
    _entityfile     = layout.LIST_META_FILE
    _entityref      = layout.META_LIST_REF

    def __init__(self, parent, list_id, altparent=None):
        """
        Initialize a new RecordList object, without metadta (yet).

        parent      is the parent entity from which the list is descended.
        list_id     the local identifier for the record list
        altparent   is a site object to search for this new entity,
                    allowing site-wide RecordType values to be found.
        """
        super(RecordList, self).__init__(parent, list_id, altparent)
        self._parent = parent
        log.debug("RecordList %s: dir %s, alt %s"%(list_id, self._entitydir, self._entityaltdir))
        return

# End.



"""
Annalist collection

A collection is represented by:
- an ID (slug)
- a URI
- a name/label
- a description
- a set of record types
- a set of list views
- a set of record views
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil
import json
from collections    import OrderedDict

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                       import layout
from annalist.exceptions            import Annalist_Error
from annalist.identifiers           import RDF, RDFS, ANNAL
# from annalist                       import util
from annalist.util                  import valid_id, extract_entity_id

from annalist.models.entity         import Entity
from annalist.models.annalistuser   import AnnalistUser
from annalist.models.recordtype     import RecordType
from annalist.models.recordview     import RecordView
from annalist.models.recordlist     import RecordList
from annalist.models.recordfield    import RecordField
from annalist.models.recordgroup    import RecordGroup
from annalist.models.recordvocab    import RecordVocab

# @@TODO: look for way to remove this view dependency from model code.  Move to RecordField 
#   module?  Hmmm... not so easy: information about render types and methods is concentrated
#   in render_utils module.  Maybe need to split render_utils functionality between
#   model and view modules?
#
from annalist.views.fields.render_utils import (
    is_render_type_literal,
    is_render_type_id,
    is_render_type_set,
    is_render_type_list,
    is_render_type_object,
    )

class Collection(Entity):

    _entitytype     = ANNAL.CURIE.Collection
    _entitytypeid   = "_collection"
    _entityview     = layout.SITE_COLL_VIEW
    _entitypath     = layout.SITE_COLL_PATH
    _entityfile     = layout.COLL_META_FILE
    _entityref      = layout.META_COLL_REF
    _contextref     = layout.COLL_CONTEXT_FILE

    def __init__(self, parentsite, coll_id):
        """
        Initialize a new Collection object, without metadta (yet).

        parentsite  is the parent site from which the new collection is descended.
        coll_id     the collection identifier for the collection
        """
        super(Collection, self).__init__(parentsite, coll_id)
        self._parentsite = parentsite
        self._types_by_id  = None
        self._types_by_uri = None
        return

    # Site

    def get_site(self):
        """
        Return site object for the site from which the current collection is accessed.
        """
        return self._parentsite

    # User permissions

    def create_user_permissions(self, user_id, user_uri,
            user_name, user_description,
            user_permissions=["VIEW"]
            ):
        user_values = (
            { ANNAL.CURIE.type:             ANNAL.CURIE.User
            , RDFS.CURIE.label:             user_name
            , RDFS.CURIE.comment:           user_description
            , ANNAL.CURIE.user_uri:         user_uri
            , ANNAL.CURIE.user_permissions: user_permissions
            })
        user = AnnalistUser.create(self, user_id, user_values)
        return user

    def get_user_permissions(self, user_id, user_uri):
        """
        Get a user permissions record (AnnalistUser).

        To return a value, both the user_id and the user_uri (typically a mailto: URI, but
        may be any *authenticated* identifier) must match.  This is to prevent access to 
        records of a deleted account being granted to a new account created with the 
        same user_id (username).

        user_id         local identifier for the type to retrieve.
        user_uri        authenticated identifier associated with the user_id.  That is,
                        the authentication service used is presumed to confirm that
                        the identifier belongs to the user currently logged in with
                        the supplied username.

        returns an AnnalistUser object for the identified user, or None.  This object contains
                information about permissions granted to the user in the current collection.
        """
        user = AnnalistUser.load(self, user_id, altparent=self._parentsite)
        log.debug("Collection.get_user_permissions: user_id %s, user_uri %s, user %r"%
            (user_id, user_uri, user)
            )
        if user:
            for f in [RDFS.CURIE.label, RDFS.CURIE.comment, ANNAL.CURIE.user_uri, ANNAL.CURIE.user_permissions]:
                if f not in user:
                    user = None
                    break
        if user and user[ANNAL.CURIE.user_uri] != user_uri:
            user = None         # URI mismatch: return None.
        return user

    # Record types

    def _update_type_cache(self, type_entity):
        """
        Add single type entity to type cache
        """
        if type_entity:
            self._types_by_id[type_entity.get_id()]   = type_entity
            self._types_by_uri[type_entity.get_uri()] = type_entity
        return

    def _flush_type(self, type_id):
        """
        Remove single identified type entity from type cache
        """
        if self._types_by_id:
            t = self._types_by_id.get(type_id, None)
            if t:
                type_uri = t.get_uri()
                self._types_by_id.pop(type_id, None)
                self._types_by_uri.pop(type_uri, None)
        return

    def _load_types(self):
        """
        Initialize cache of RecordType entities
        """
        if not (self._types_by_id and self._types_by_uri):
            self._types_by_id  = {}
            self._types_by_uri = {}
            for type_id in self._children(RecordType, altparent=self._parentsite):
                t = RecordType.load(self, type_id, altparent=self._parentsite)
                self._update_type_cache(t)
        return

    def types(self, include_alt=True):
        """
        Generator enumerates and returns record types that may be stored
        """
        altparent = self._parentsite if include_alt else None
        for f in self._children(RecordType, altparent=altparent):
            t = self.get_type(f)
            if t and t.get_id() != "_initial_values":
                yield t
        return

    def add_type(self, type_id, type_meta):
        """
        Add a new record type to the current collection

        type_id     identifier for the new type, as a string
                    with a form that is valid as URI path segment.
        type_meta   a dictionary providing additional information about
                    the type to be created.

        Returns a RecordType object for the newly created type.
        """
        t = RecordType.create(self, type_id, type_meta)
        if self._types_by_id:
            self._update_type_cache(t)
        return t

    def get_type(self, type_id):
        """
        Retrieve identified type description

        type_id     local identifier for the type to retrieve.

        returns a RecordType object for the identified type, or None.
        """
        self._load_types()
        t = self._types_by_id.get(type_id, None)
        # Was it created but not cached?
        if not t and RecordType.exists(self, type_id, altparent=self._parentsite):
            t = RecordType.load(self, type_id, altparent=self._parentsite)
            self._update_type_cache(t)
        return t

    def get_uri_type(self, type_uri, include_alt=True):
        """
        Return type entity corresponding to the supplied type URI
        """
        self._load_types()
        t = self._types_by_uri.get(type_uri, None)
        return t

    def remove_type(self, type_id):
        """
        Remove identified type description

        type_id     local identifier for the type to remove.

        Returns a non-False status code if the type is not removed.
        """
        self._flush_type(type_id)
        s = RecordType.remove(self, type_id)
        return s

    # Record views

    def views(self, include_alt=True):
        """
        Generator enumerates and returns record views that may be stored
        """
        altparent = self._parentsite if include_alt else None
        for f in self._children(RecordView, altparent=altparent):
            v = self.get_view(f)
            if v and v.get_id() != "_initial_values":
                yield v
        return

    def add_view(self, view_id, view_meta):
        """
        Add a new record view to the current collection

        view_id     identifier for the new view, as a string
                    with a form that is valid as URI path segment.
        view_meta   a dictionary providing additional information about
                    the view to be created.

        returns a RecordView object for the newly created view.
        """
        v = RecordView.create(self, view_id, view_meta)
        return v

    def get_view(self, view_id):
        """
        Retrieve identified view description

        view_id     local identifier for the view to retrieve.

        returns a RecordView object for the identified view, or None.
        """
        v = RecordView.load(self, view_id, altparent=self._parentsite)
        return v

    def remove_view(self, view_id):
        """
        Remove identified view description

        view_id     local identifier for the view to remove.

        Returns a non-False status code if the view is not removed.
        """
        s = RecordView.remove(self, view_id)
        return s

    # Record lists

    def lists(self, include_alt=True):
        """
        Generator enumerates and returns record lists that may be stored
        """
        altparent = self._parentsite if include_alt else None
        for f in self._children(RecordList, altparent=altparent):
            l = self.get_list(f)
            if l and l.get_id() != "_initial_values":
                yield l
        return

    def add_list(self, list_id, list_meta):
        """
        Add a new record list to the current collection

        list_id     identifier for the new list, as a string
                    with a form that is valid as URI path segment.
        list_meta   a dictionary providing additional information about
                    the list to be created.

        returns a RecordList object for the newly created list.
        """
        l = RecordList.create(self, list_id, list_meta)
        return l

    def get_list(self, list_id):
        """
        Retrieve identified list description

        list_id     local identifier for the list to retrieve.

        returns a RecordList object for the identified list, or None.
        """
        l = RecordList.load(self, list_id, altparent=self._parentsite)
        return l

    def remove_list(self, list_id):
        """
        Remove identified list description

        list_id     local identifier for the list to remove.

        Returns a non-False status code if the list is not removed.
        """
        s = RecordList.remove(self, list_id)
        return s

    def set_default_list(self, list_id):
        """
        Set and save the default list to be displayed for the current collection.
        """
        self[ANNAL.CURIE.default_list] = list_id
        self._save()
        return

    def get_default_list(self):
        """
        Return the default list to be displayed for the current collection.
        """
        list_id = self.get(ANNAL.CURIE.default_list, None)
        if list_id and not RecordList.exists(self, list_id, altparent=self._parentsite):
            log.warning(
                "Default list %s for collection %s does not exist"%
                (list_id, self.get_id())
                )
            list_id = None
        return list_id 

    # JSON-LD context data

    def generate_coll_jsonld_context(self):
        """
        (Re)generate JSON-LD context description for the current collection.

        get_field_uri_context
                is a supplied function that accepts a RecordField object abnd
                returns a context dictionary for the field thus described.
        """
        # Build context data
        context = self.get_coll_jsonld_context()
        # Assemble and write out context description
        with self._metaobj(
                layout.COLL_META_CONTEXT_PATH,
                layout.COLL_CONTEXT_FILE,
                "wt"
                ) as context_io:
            json.dump(
                { "@context": context }, 
                context_io, indent=2, separators=(',', ': ')
                )
        return

    def get_coll_jsonld_context(self):
        """
        Return dictionary containing context structure for collection.
        """
        # @@REVIEW: as a workaround for a problem with @base handling in rdflib-jsonld, don't
        #           include @base in context.
        #
        # context           = OrderedDict(
        #     { "@base":          self.get_url() + layout.COLL_CONTEXT_PATH
        #     , ANNAL.CURIE.type: { "@type": "@id" }
        #     })
        #
        # Use OrderedDict to allow some control over ordering of context file contents:
        # this is for humane purposes only, and is not technically important.
        context           = OrderedDict(
            { ANNAL.CURIE.type: { "@type": "@id" }
            })
        # Common import/upload fields
        context.update(
            { 'resource_name': "annal:resource_name"
            , 'resource_type': "annal:resource_type"
            })
        # upload-file fields
        context.update(
            { 'upload_name':   "annal:upload_name"
            , 'uploaded_file': "annal:uploaded_file"
            , 'uploaded_size': "annal:uploaded_size"
            })
        # import-resource fields
        context.update(
            { 'import_name':   "annal:import_name"
            , 'import_url':    
              { "@id": "annal:import_url"
              , "@type": "@id"
              }
            })
        # Scan vocabs, generate prefix data
        for v in self.child_entities(RecordVocab, altparent=self._parentsite):
            vid = v.get_id()
            if vid != "_initial_values":
                context[v.get_id()] = v[ANNAL.CURIE.uri]
        # Scan view fields and generate context data for property URIs used
        for v in self.child_entities(RecordView, altparent=self._parentsite):
            for fref in v[ANNAL.CURIE.view_fields]:
                fid  = extract_entity_id(fref[ANNAL.CURIE.field_id])
                vuri = fref.get(ANNAL.CURIE.property_uri, None)
                furi, fcontext = self.get_field_uri_jsonld_context(fid, self.get_field_jsonld_context)
                # fcontext['vid'] = v.get_id()
                # fcontext['fid'] = fid
                self.set_field_uri_jsonld_context(vuri or furi, fcontext, context)
        # Scan group fields and generate context data for property URIs used
        for g in self.child_entities(RecordGroup, altparent=self._parentsite):
            for gref in g[ANNAL.CURIE.group_fields]:
                fid  = extract_entity_id(gref[ANNAL.CURIE.field_id])
                guri = gref.get(ANNAL.CURIE.property_uri, None)
                furi, fcontext = self.get_field_uri_jsonld_context(fid, self.get_field_jsonld_context)
                # fcontext['gid'] = g.get_id()
                # fcontext['fid'] = fid
                self.set_field_uri_jsonld_context(guri or furi, fcontext, context)
        return context

    def get_field_uri_jsonld_context(self, fid, get_field_context):
        """
        Access field description, and return field property URI and appropriate 
        property description for JSON-LD context.

        If there is no corresponding field description, returns (None, None)

        If no context should be generated for the field URI, returns (uri, None)
        """
        f = RecordField.load(self, fid, altparent=self._parentsite)
        if f is None:
            return (None, None)
        return (f[ANNAL.CURIE.property_uri], get_field_context(f))

    def set_field_uri_jsonld_context(self, puri, fcontext, property_contexts):
        """
        Save property context description into supplied property_contexts dictionary.  
        If the context is already defined, generate warning if there is a compatibility 
        problem.
        """
        if puri:
            uri_parts = puri.split(":")
            if len(uri_parts) > 1:
                # Ignore URIs without ':'
                if puri in property_contexts:
                    pcontext = property_contexts[puri]
                    if ( ( not fcontext ) or
                         ( pcontext.get("@type", None)      != fcontext.get("@type", None) ) or
                         ( pcontext.get("@container", None) != fcontext.get("@container", None) ) ):
                        log.warning(
                            "Incompatible use of property %s (%r, %r)"% (puri, fcontext, pcontext)
                            )
                elif ( fcontext and
                       ( uri_parts[0] in property_contexts ) or         # Prefix defined vocab?
                       ( uri_parts[0] in ["http", "https", "file"] ) ): # Full URI?
                    property_contexts[puri] = fcontext
        return

    # @@TODO: move this away from model logic, as it represents a dependency on view logic?
    @staticmethod
    def get_field_jsonld_context(fdesc):
        """
        Returns a context description for the supplied field description.

        Returns None if no property context information is needed for the 
        supplied field.
        """
        rtype = extract_entity_id(fdesc[ANNAL.CURIE.field_render_type])
        vmode = extract_entity_id(fdesc[ANNAL.CURIE.field_value_mode])
        if vmode in ["Value_entity", "Value_field"]:
            rtype = "Enum"
        elif vmode == "Value_import":
            rtype = "URIImport"
        elif vmode == "Value_upload":
            rtype = "FileUpload"
        if is_render_type_literal(rtype):
            fcontext = None # { "@type": "xsd:string" }
        elif is_render_type_id(rtype):
            fcontext = { "@type": "@id" }   # Add type from field descr?
        elif is_render_type_set(rtype):
            fcontext = { "@container": "@set"}
        elif is_render_type_list(rtype):
            fcontext = { "@container": "@list"}
        elif is_render_type_object(rtype):
            fcontext = None
        else:
            raise ValueError("Unexpected value mode or render type (%s, %s)"%(vmode, rtype))
        return fcontext

# End.



"""
Tests for file upload functions.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2015, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.contrib.auth.models     import User
from django.test                    import TestCase
from django.core.urlresolvers       import resolve, reverse
from django.test.client             import Client

from annalist                       import util
from annalist.identifiers           import ANNAL
from annalist.models.entity         import EntityRoot, Entity
from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordlist     import RecordList
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.entitytypeinfo import EntityTypeInfo

from tests                  import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from init_tests             import init_annalist_test_site, init_annalist_test_coll, resetSitedata
from AnnalistTestCase       import AnnalistTestCase
from entity_testutils       import (
    create_test_user,
    create_user_permissions,
    context_list_entities,
    context_list_head_fields,
    context_list_item_fields
    )
from entity_testentitydata  import (
    entity_url, entitydata_edit_url, 
    entitydata_default_view_context_data, entitydata_default_view_form_data,
    )

#   -----------------------------------------------------------------------------
#
#   Test data
#
#   -----------------------------------------------------------------------------

test_upload_type_create_values = (
    { 'annal:type':                 "annal:Type"
    , 'rdfs:label':                 "test_upload_type label"
    , 'rdfs:comment':               "test_upload_type comment"
    , 'annal:uri':                  "test:type/test_upload_type"
    , 'annal:type_view':            "test_upload_view"
    , 'annal:type_list':            "test_upload_list"
    })

test_reference_type_create_values = (
    { 'annal:type':                 "annal:Type"
    , 'rdfs:label':                 "test_reference_type label"
    , 'rdfs:comment':               "test_reference_type comment"
    , 'annal:uri':                  "test:type/test_reference_type"
    , 'annal:type_view':            "test_reference_view"
    , 'annal:type_list':            "test_reference_list"
    })

test_image_type_create_values = (
    { 'annal:type':                 "annal:Type"
    , 'rdfs:label':                 "test_image_type label"
    , 'rdfs:comment':               "test_image_type comment"
    , 'annal:uri':                  "test:type/test_image_type"
    , 'annal:type_view':            "test_image_view"
    , 'annal:type_list':            "test_image_list"
    })

test_upload_file_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_upload_view label"
    , 'rdfs:comment':               "test_upload_view comment"
    , 'annal:record_type':          ""
    , 'annal:add_field':            "yes"
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_upload_file"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_upload_image_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_upload_image_view label"
    , 'rdfs:comment':               "test_upload_image_view comment"
    , 'annal:record_type':          ""
    , 'annal:add_field':            "yes"
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_upload_image"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_reference_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_reference_view label"
    , 'rdfs:comment':               "test_reference_view comment"
    , 'annal:record_type':          ""
    , 'annal:add_field':            "yes"
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_reference"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_image_ref_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_image_view label"
    , 'rdfs:comment':               "test_image_view comment"
    , 'annal:record_type':          ""
    , 'annal:add_field':            "yes"
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_image_ref"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_image_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_image_view label"
    , 'rdfs:comment':               "test_image_view comment"
    , 'annal:record_type':          ""
    , 'annal:add_field':            "yes"
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_image"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_upload_file_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "upl_field"
    , 'rdfs:label':                     "test_upload_file_field label"
    , 'rdfs:comment':                   "test_upload_file_field comment"
    , 'annal:property_uri':             "test:upload"
    , 'annal:field_render_type':        "FileUpload"
    , 'annal:field_value_mode':         "Value_upload"
    , 'annal:field_target_type':        "annal:Richtext"
    , 'annal:placeholder':              "(File to upload)"
    , 'annal:default_value':            ""
    })

test_upload_image_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "upl_field"
    , 'rdfs:label':                     "test_upload_image_field label"
    , 'rdfs:comment':                   "test_upload_image_field comment"
    , 'annal:property_uri':             "test:upload"
    , 'annal:field_render_type':        "FileUpload"
    , 'annal:field_value_mode':         "Value_upload"
    , 'annal:field_target_type':        "annal:Image"
    , 'annal:placeholder':              "(Image to upload)"
    , 'annal:default_value':            ""
    })

test_reference_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "ref_field"
    , 'rdfs:label':                     "test_reference_field label"
    , 'rdfs:comment':                   "test_reference_field comment"
    , 'annal:property_uri':             "test:reference"
    , 'annal:field_render_type':        "URILink"
    , 'annal:field_value_mode':         "Value_field"
    , 'annal:field_target_type':        "annal:Identifier"
    , 'annal:field_ref_type':           "testupltype"
    , 'annal:field_ref_restriction':    "ALL"
    , 'annal:field_ref_field':          "test:upload"
    , 'annal:placeholder':              "(Uploaded file entity reference)"
    , 'annal:default_value':            ""
    })

test_image_ref_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "ref_image"
    , 'rdfs:label':                     "test_image_ref_field label"
    , 'rdfs:comment':                   "test_image_ref_field comment"
    , 'annal:property_uri':             "test:reference"
    , 'annal:field_render_type':        "RefImage"
    , 'annal:field_value_mode':         "Value_field"
    , 'annal:field_target_type':        "annal:Identifier"
    , 'annal:field_ref_type':           "testupltype"
    , 'annal:field_ref_restriction':    "ALL"
    , 'annal:field_ref_field':          "test:upload"
    , 'annal:placeholder':              "(Uploaded image reference)"
    , 'annal:default_value':            ""
    })

test_image_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "img_field"
    , 'rdfs:label':                     "test_image_field label"
    , 'rdfs:comment':                   "test_image_field comment"
    , 'annal:property_uri':             "test:image"
    , 'annal:field_render_type':        "RefImage"
    , 'annal:field_value_mode':         "Value_upload"
    , 'annal:field_target_type':        "annal:Image"
    , 'annal:placeholder':              "(Image to upload)"
    , 'annal:default_value':            ""
    })

def test_upload_file_field_value():
    return (
        { "resource_name":              "upl_field.md"
        , "resource_type":              "text/markdown"
        , "upload_name":                "upl_field"
        , "uploaded_size":              1005
        , "uploaded_file":              "README.md"
        })

def test_upload_image_field_value():
    return (
        { "resource_name":              "upl_field.jpg"
        , "resource_type":              "image/jpeg"
        , "upload_name":                "upl_field"
        , "uploaded_size":              1547926
        , "uploaded_file":              "test-image.jpg"
        })

def test_image_field_value():
    return (
        { "resource_name":              "img_field.jpg"
        , "resource_type":              "image/jpeg"
        , "upload_name":                "img_field"
        , "uploaded_size":              1547926
        , "uploaded_file":              "test-image.jpg"
        })

def test_imp_entity_create_values(entity_id):
    return (
        { 'rdfs:label':                 "test_imp_entity %s label"%entity_id
        , 'rdfs:comment':               "test_imp_entity %s comment"%entity_id
        , 'test:upload':                test_upload_file_field_value()
        })

def test_ref_entity_create_values(entity_id):
    return (
        { 'rdfs:label':                 "test_ref_entity %s label"%entity_id
        , 'rdfs:comment':               "test_ref_entity %s comment"%entity_id
        , 'test:reference':             "testupltype/"+entity_id
        })

def test_img_entity_create_values(entity_id):
    return (
        { 'rdfs:label':                 "test_img_entity %s label"%entity_id
        , 'rdfs:comment':               "test_img_entity %s comment"%entity_id
        , 'test:image':                 test_image_field_value()
        })

#   -----------------------------------------------------------------------------
#
#   Linked record tests
#
#   -----------------------------------------------------------------------------

class UploadResourceTest(AnnalistTestCase):
    """
    Tests for resource import
    """

    def setUp(self):
        self.filepath  = "%s/README.md"%TestBaseDir
        self.fileuri   = "file://"+self.filepath
        self.imagepath = "%s/test-image.jpg"%TestBaseDir
        self.imageuri  = "file://"+self.imagepath
        init_annalist_test_site()
        init_annalist_test_coll()
        self.testsite  = Site(TestBaseUri, TestBaseDir)
        self.testcoll  = Collection(self.testsite, "testcoll")
        # Populate collection with linked record types, views and lists

        # Types
        self.test_upl_type = RecordType.create(
            self.testcoll, "testupltype", test_upload_type_create_values
            )
        self.test_ref_type = RecordType.create(
            self.testcoll, "testreftype", test_reference_type_create_values
            )
        self.test_image_type = RecordType.create(
            self.testcoll, "testimgtype", test_image_type_create_values
            )

        # Views
        self.test_upl_file_view = RecordView.create(
            self.testcoll, "testuplfileview", test_upload_file_view_create_values
            )
        self.test_upl_image_view = RecordView.create(
            self.testcoll, "testuplimageview", test_upload_image_view_create_values
            )
        self.test_ref_file_view = RecordView.create(
            self.testcoll, "testrefview", test_reference_view_create_values
            )
        self.test_ref_image_view = RecordView.create(
            self.testcoll, "testimgrefview", test_image_ref_view_create_values
            )
        self.test_image_view = RecordView.create(
            self.testcoll, "testimgview", test_image_view_create_values
            )

        # Fields
        self.test_upl_file_field = RecordField.create(
            self.testcoll, "Test_upload_file", test_upload_file_field_create_values
            )
        self.test_upl_image_field = RecordField.create(
            self.testcoll, "Test_upload_image", test_upload_image_field_create_values
            )
        self.test_ref_file_field = RecordField.create(
            self.testcoll, "Test_reference", test_reference_field_create_values
            )
        self.test_ref_image_field = RecordField.create(
            self.testcoll, "Test_image_ref", test_image_ref_field_create_values
            )
        self.test_image_field = RecordField.create(
            self.testcoll, "Test_image", test_image_field_create_values
            )

        # Create data records for testing import and references:
        test_entity_ids = ("test1", "test2")
        test_entity_ids = ("test1",)
        self.test_upl_type_info = EntityTypeInfo(
            self.testsite, self.testcoll, "testupltype", create_typedata=True
            )
        for entity_id in test_entity_ids:
            self.test_upl_type_info.create_entity(
                entity_id, test_imp_entity_create_values(entity_id)
                )
        self.test_ref_type_info = EntityTypeInfo(
            self.testsite, self.testcoll, "testreftype", create_typedata=True
            )
        for entity_id in test_entity_ids:
            self.test_ref_type_info.create_entity(
                entity_id, test_ref_entity_create_values(entity_id)
                )
        self.test_img_type_info = EntityTypeInfo(
            self.testsite, self.testcoll, "testimgtype", create_typedata=True
            )
        for entity_id in test_entity_ids:
            self.test_img_type_info.create_entity(
                entity_id, test_img_entity_create_values(entity_id)
                )
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    # Utility functions

    # Tests

    def test_entity_fileobj(self):
        test1    = self.test_upl_type_info.get_entity("test1")
        test1dir, test1file = test1._dir_path()
        testobj1 = self.test_upl_type_info.get_fileobj(
            "test1", "test1res", "annal:Text", "text/plain", "wb"
            )
        testobj1.write("Test data test1res.txt")
        self.assertEqual(testobj1.name, test1dir+"/test1res.txt")
        testobj1.close()
        testobj2 = self.test_upl_type_info.get_fileobj(
            "test1", "test1res", "annal:Text", "text/plain", "rb"
            )
        self.assertEqual(testobj2.read(), "Test data test1res.txt")
        testobj2.close()
        return

    def test_util_fileobj(self):
        resource_fileobj, resource_url, resource_type = util.open_url(self.fileuri)
        self.assertEqual(resource_url,  self.fileuri)
        self.assertEqual(resource_type, "text/markdown")
        testobj1 = self.test_upl_type_info.get_fileobj(
            "test1", "test1res", "annal:Richtext", resource_type, "wb"
            )
        util.copy_resource_to_fileobj(resource_fileobj, testobj1)
        resource_fileobj.close()
        testobj1.close()
        # Read back both and compare
        siteobj = open(TestBaseDir+"/README.md", "rb")
        testobj = self.test_upl_type_info.get_fileobj(
            "test1", "test1res", "annal:Richtext", resource_type, "rb"
            )
        self.assertEqual(siteobj.read(), testobj.read())
        return

    def test_upload_file_resource(self):
        # See https://docs.djangoproject.com/en/1.7/topics/testing/tools/#django.test.Client.post
        with open(self.filepath) as fp:
            f = entitydata_default_view_form_data(entity_id="test1", action="edit")
            f['upl_field'] = fp     # Upload file with submission
            u = entitydata_edit_url("edit", "testcoll", "testupltype", entity_id="test1", view_id="testuplfileview")
            r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        # Retrieve updated form
        r = self.client.get(u)
        # Test context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        i = 3
        self.assertEqual(r.context['fields'][i].field_id,     "Test_upload_file")
        self.assertDictionaryMatch(r.context['fields'][i].field_value, test_upload_file_field_value())
        # Read back and compare entity resource just created
        siteobj = open(self.filepath, "rb")
        testobj = self.test_upl_type_info.get_fileobj(
            "test1", "upl_field", "annal:Richtext", "text/markdown", "rb"
            )
        self.assertEqual(siteobj.read(), testobj.read())
        return

    def test_reference_uploaded_resource(self):
        # Create uploaded resource (see previous test)
        with open(self.filepath) as fp:
            f = entitydata_default_view_form_data(entity_id="test1", action="edit")
            f['upl_field'] = fp     # Upload file with submission
            u = entitydata_edit_url(
                "view", "testcoll", "testupltype", entity_id="test1", view_id="testuplfileview"
                )
            r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        # Display resource with reference
        u = entitydata_edit_url(
            "view", "testcoll", "testreftype", entity_id="test1", view_id="testrefview"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check display context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        self.assertEqual(r.context['fields'][i].field_value,  "test_ref_entity test1 label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        self.assertEqual(r.context['fields'][i].field_value,  "test_ref_entity test1 comment")
        i = 3
        basepath = TestBasePath + "/c/testcoll/d/testupltype/"
        # print "\n*****\n"+repr(r.context['fields'][i].target_value)+"\n*****\n"
        self.assertEqual(r.context['fields'][i].field_id,     "Test_reference")
        self.assertEqual(r.context['fields'][i].field_value,        "testupltype/test1")
        self.assertEqual(r.context['fields'][i].field_value_link,   basepath+"test1/")
        self.assertEqual(r.context['fields'][i].target_value['upload_name'],   "upl_field")
        self.assertEqual(r.context['fields'][i].target_value['resource_name'], "upl_field.md")
        self.assertEqual(r.context['fields'][i].target_value['resource_type'], "text/markdown")
        self.assertEqual(r.context['fields'][i].target_value['uploaded_file'], "README.md")
        self.assertEqual(r.context['fields'][i].target_value['uploaded_size'], 1005)
        self.assertEqual(r.context['fields'][i].target_value_link,  basepath+"test1/upl_field.md")
        return

    def test_upload_image_resource(self):
        # See https://docs.djangoproject.com/en/1.7/topics/testing/tools/#django.test.Client.post
        with open(self.imagepath) as fp:
            f = entitydata_default_view_form_data(entity_id="test1", action="edit")
            f['upl_field'] = fp     # Upload file with submission
            u = entitydata_edit_url("edit", "testcoll", "testupltype", entity_id="test1", view_id="testuplimageview")
            r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        # Retrieve updated form
        r = self.client.get(u)
        # Test context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        i = 3
        self.assertEqual(r.context['fields'][i].field_id,     "Test_upload_image")
        self.assertDictionaryMatch(r.context['fields'][i].field_value, test_upload_image_field_value())
        # Read back and compare entity resource just created
        siteobj = open(self.imagepath, "rb")
        testobj = self.test_upl_type_info.get_fileobj(
            "test1", "upl_field", "annal:Image", "image/jpeg", "rb"
            )
        self.assertTrue(siteobj.read() == testobj.read(), "Uploaded image != original")
        # self.assertEqual(siteobj.read(), testobj.read())
        return

    def test_reference_uploaded_image(self):
        self.test_upload_image_resource()

        # Display resource with image reference
        u = entitydata_edit_url("view", "testcoll", "testreftype", entity_id="test1", view_id="testimgrefview")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check display context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        self.assertEqual(r.context['fields'][i].field_value,  "test_ref_entity test1 label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        self.assertEqual(r.context['fields'][i].field_value,  "test_ref_entity test1 comment")
        i = 3
        basepath = TestBasePath + "/c/testcoll/d/testupltype/"
        # print "\n*****\n"+repr(r.context['fields'][i].target_value)+"\n*****\n"
        self.assertEqual(r.context['fields'][i].field_id,     "Test_image_ref")
        self.assertEqual(r.context['fields'][i].field_value,        "testupltype/test1")
        self.assertEqual(r.context['fields'][i].field_value_link,   basepath+"test1/")
        self.assertEqual(r.context['fields'][i].target_value['upload_name'],   "upl_field")
        self.assertEqual(r.context['fields'][i].target_value['resource_name'], "upl_field.jpg")
        self.assertEqual(r.context['fields'][i].target_value['resource_type'], "image/jpeg")
        self.assertEqual(r.context['fields'][i].target_value['uploaded_file'], "test-image.jpg")
        self.assertEqual(r.context['fields'][i].target_value['uploaded_size'], 1547926)
        self.assertEqual(r.context['fields'][i].target_value_link,  basepath+"test1/upl_field.jpg")
        # Check for rendered image link
        # log.info(r.content)
        field_details = (
            { "basepath":   TestBasePath
            , "coll_id":    "testcoll"
            , "type_id":    "testupltype"
            , "entity_id":  "test1"
            , "field_id":   "upl_field"
            })
        img_element = (
            """<div class="small-12 columns"> """+
              """<div class="row view-value-row"> """+
                """<div class="view-label small-12 medium-2 columns"> """+
                  """<span>test_image_ref_field label</span> """+
                """</div> """+
                """<div class="view-value small-12 medium-10 columns"> """+
                  """<a href="%(basepath)s/c/%(coll_id)s/d/%(type_id)s/%(entity_id)s/%(field_id)s.jpg" target="_blank"> """+
                    """<img src="%(basepath)s/c/%(coll_id)s/d/%(type_id)s/%(entity_id)s/%(field_id)s.jpg" """+
                    """     alt="Image at '%(basepath)s/c/%(coll_id)s/d/%(type_id)s/%(entity_id)s/%(field_id)s.jpg'" /> """+
                  """</a> """+
                """</div> """+
              """</div> """+
            """</div> """
            )%field_details
        self.assertContains(r, img_element, html=True)
        return

    def test_image_edit_field(self):
        # Upload to an image view field
        with open(self.imagepath) as fp:
            f = entitydata_default_view_form_data(entity_id="test1", action="edit")
            f['img_field'] = fp     # Upload file with submission
            u = entitydata_edit_url("edit", "testcoll", "testimgtype", entity_id="test1", view_id="testimgview")
            r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")

        # Read back and compare entity resource just created
        siteobj = open(self.imagepath, "rb")
        testobj = self.test_img_type_info.get_fileobj(
            "test1", "img_field", "annal:Image", "image/jpeg", "rb"
            )
        self.assertTrue(siteobj.read() == testobj.read(), "Referenced image != original")

        # Retrieve updated form
        r = self.client.get(u)
        # Test context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        i = 3
        self.assertEqual(r.context['fields'][i].field_id,     "Test_image")
        self.assertDictionaryMatch(r.context['fields'][i].field_value, test_image_field_value())
        return

    def test_image_view_field(self):
        # This test is for an image field that supports file upload in the same entity

        # Upload image
        self.test_image_edit_field()

        # Display resource in view mode
        u = entitydata_edit_url("view", "testcoll", "testimgtype", entity_id="test1", view_id="testimgview")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check display context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        i = 3
        basepath = TestBasePath + "/c/testcoll/d/testimgtype/"
        # print "\n*****\n"+repr(r.context['fields'][i].target_value)+"\n*****\n"
        self.assertEqual(r.context['fields'][i].field_id,           "Test_image")
        self.assertEqual(r.context['fields'][i].field_value,        test_image_field_value())
        self.assertEqual(r.context['fields'][i].target_value_link,  basepath+"test1/img_field.jpg")
        # Check for rendered image link
        # log.info(r.content)
        field_details = (
            { "basepath":   TestBasePath
            , "coll_id":    "testcoll"
            , "type_id":    "testimgtype"
            , "entity_id":  "test1"
            , "field_id":   "img_field"
            })
        img_element = (
            """<div class="small-12 columns"> """+
              """<div class="row view-value-row"> """+
                """<div class="view-label small-12 medium-2 columns"> """+
                  """<span>test_image_field label</span> """+
                """</div> """+
                """<div class="view-value small-12 medium-10 columns"> """+
                  """<a href="%(basepath)s/c/%(coll_id)s/d/%(type_id)s/%(entity_id)s/%(field_id)s.jpg" target="_blank"> """+
                    """<img src="%(basepath)s/c/%(coll_id)s/d/%(type_id)s/%(entity_id)s/%(field_id)s.jpg" """+
                    """     alt="Image at '%(basepath)s/c/%(coll_id)s/d/%(type_id)s/%(entity_id)s/%(field_id)s.jpg'" /> """+
                  """</a> """+
                """</div> """+
              """</div> """+
            """</div> """
            )%field_details
        self.assertContains(r, img_element, html=True)
        return

    def test_image_edit(self):
        # This test that entity editing leaves attachment intact

        # Upload image
        self.test_image_edit_field()

        # Edit entity
        f = entitydata_default_view_form_data(entity_id="test1", action="edit", update="Updated")
        u = entitydata_edit_url("edit", "testcoll", "testimgtype", entity_id="test1", view_id="testimgview")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")

        # Retrieve updated form
        r = self.client.get(u)
        # Test context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        self.assertEqual(r.context['fields'][i].field_value,  "Updated testcoll/testtype/test1")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        i = 3
        self.assertEqual(r.context['fields'][i].field_id,     "Test_image")
        self.assertDictionaryMatch(r.context['fields'][i].field_value, test_image_field_value())

        # Read back and compare entity resource
        siteobj = open(self.imagepath, "rb")
        testobj = self.test_img_type_info.get_fileobj(
            "test1", "img_field", "annal:Image", "image/jpeg", "rb"
            )
        self.assertTrue(siteobj.read() == testobj.read(), "Edited entity image != original")
        return

    def test_image_rename(self):
        # This test that entity renaming also copies over an attachment

        # Upload image
        self.test_image_edit_field()

        # Rename entity
        f = entitydata_default_view_form_data(orig_id="test1", entity_id="test_new", action="edit")
        u = entitydata_edit_url("edit", "testcoll", "testimgtype", entity_id="test1", view_id="testimgview")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")

        # Read back and compare renamed entity resource
        siteobj = open(self.imagepath, "rb")
        testobj = self.test_img_type_info.get_fileobj(
            "test_new", "img_field", "annal:Image", "image/jpeg", "rb"
            )
        self.assertTrue(siteobj.read() == testobj.read(), "Renamed entity image != original")
        return

# End.



"""
Renderer and value mapper for text values selected from a list of options.
In some cases, the ren dered edit control also inclused a button for 
creating a new value.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import traceback
import logging
log = logging.getLogger(__name__)

from annalist               import message
from annalist.exceptions    import TargetIdNotFound_Error, TargetEntityNotFound_Error
from annalist.util          import fill_type_entity_id

from annalist.views.fields.render_base          import RenderBase
from annalist.views.fields.render_fieldvalue    import (
    RenderFieldValue,
    get_field_edit_value,
    get_field_view_value
    )
from annalist.views.form_utils.fieldchoice      import FieldChoice

from django.template        import Template, Context

#   ----------------------------------------------------------------------------
#
#   Select value templates
#
#   ----------------------------------------------------------------------------

# @@TODO: remove references to field value and special case code in bound_field.
#         the continuation URI will need to be provided separately in the context
#         and mentioned separately in the templates

edit_options = (
    '''{% for opt in field_options %} '''+
      '''{% if opt.value == encoded_field_value %} '''+
        '''{% if opt.value == "" %} '''+
          '''<option value="" selected="selected">{{field.field_placeholder}}</option>\n'''+
        '''{% else %} '''+
          '''<option value="{{opt.value}}" selected="selected">{{opt.label}}&nbsp;&nbsp;&nbsp;({{opt.value}})</option>\n'''+
        '''{% endif %} '''+
      '''{% else %} '''+
        '''{% if opt.value == "" %} '''+
          '''<option value="">{{field.field_placeholder}}</option>\n'''+
        '''{% else %} '''+
          '''<option value="{{opt.value}}">{{opt.label}}&nbsp;&nbsp;&nbsp;({{opt.value}})</option>\n'''+
        '''{% endif %} '''+
      '''{% endif %} '''+
    '''{% endfor %} '''
    )

view_select = (
    """<!-- fields.render_select.view_select -->
    {% if field_linkval %}
      <a href="{{field_linkval}}">{{field_labelval}}</a>
    {% elif field_textval and field_textval != "" %}
      <span class="value-missing">{{field_labelval}}</span>
    {% else %}
    <span class="value-blank">"""+
      message.NO_SELECTION%{'id': "{{field.field_label}}"}+
    """</span>
    {% endif %}
    """)

edit_select = (
    """<!-- fields.render_select.edit_select -->
    <div class="row"> 
      <div class="small-10 columns view-value less-new-button">
        <select name="{{repeat_prefix}}{{field.field_name}}">
    """+
    edit_options+
    """
        </select>
      </div>
      <div class="small-2 columns view-value new-button left small-text-right">
        <button type="submit" 
                name="{{repeat_prefix}}{{field.field_name}}__new_edit" 
                value="New"
                title="Define new {{field.field_label}}"
        >
          <span class="select-edit-button-text">+&#x270D;</span>
        </button>
      </div>
    </div>
    """)

view_choice = (
    """<!-- fields.render_select.view_choice -->
    {% if field_linkval %}
      <a href="{{field_linkval}}">{{field_labelval}}</a>
    {% elif field_textval and field_textval != "" %}
      <span class="value-missing">{{field_labelval}}</span>
    {% else %}
    <span class="value-blank">"""+
      message.NO_SELECTION%{'id': "{{field.field_label}}"}+
    """</span>
    {% endif %}
    """)

edit_choice = (
    """<!-- fields.render_select.edit_choice -->
    <select name="{{repeat_prefix}}{{field.field_name}}">
    """+
    edit_options+
    """
    </select>
    """)

view_entitytype = (
    """<!-- fields.render_select.view_entitytype -->
    {% if field_linkval %}
      <a href="{{field_linkval}}">{{field_labelval}}</a>
    {% elif field_textval and field_textval != "" %}
      <span class="value-missing">{{field_labelval}}</span>
    {% else %}
    <span class="value-blank">"""+
      message.NO_SELECTION%{'id': "{{field.field_label}}"}+
    """</span>
    {% endif %}
    """)

edit_entitytype = (
    # Note use of fixed field name
    """<!-- fields.render_select.edit_entitytype -->
    <select name="entity_type">
    """+
    edit_options+
    """
    </select>
    """)

view_view_choice = (
    """<!-- field/annalist_view_view_choice.html -->
    <span>{{field.field_value}}</span>
    """)

edit_view_choice = (
    """<!-- field/annalist_edit_view_choice.html -->
    <div class="row">
      <div class="small-9 columns">
        <select name="{{repeat_prefix}}{{field.field_name}}">
        """+
        edit_options+
        """
        </select>
      </div>
      <div class="small-3 columns">
        <input type="submit" name="use_view" value="Show view" />
      </div>
    </div>
    """)

#   ----------------------------------------------------------------------------
#
#   Select text value mapping
#
#   ----------------------------------------------------------------------------

class SelectValueMapper(RenderBase):
    """
    Value mapper class for text selected from a list of choices.
    """

    @classmethod
    def encode(cls, data_value):
        """
        Encodes supplied data value as an option value to be selected in a 
        <select> form input.
        """
        return data_value or ""

#   ----------------------------------------------------------------------------
#
#   Select text field renderers
#
#   ----------------------------------------------------------------------------

class Select_view_renderer(object):
    """
    Render select value for viewing using supplied template

    """
    def __init__(self, template):
        self._template = Template(template)
        return

    def render(self, context):
        try:
            # val      = get_field_view_value(context, None)
            val      = get_field_edit_value(context, None)
            typval   = fill_type_entity_id(val, context['field']['field_ref_type'])
            textval  = SelectValueMapper.encode(typval)
            labelval = textval
            linkval  = None
            linkcont = context['field']['continuation_param']
            options  = context['field']['options']
            for o in options:
                # log.info("Select_view_renderer.render: option %r"%(o,))
                if textval == o.value:
                    labelval = o.label
                    linkval  = o.link
                    break
            # log.info(
            #     "Select_view_renderer.render: textval %s, labelval %s, linkval %s"%
            #     (textval, labelval, linkval)
            #     )
        except TargetIdNotFound_Error as e:
            log.debug(repr(e))
            textval = ""
        except TargetEntityNotFound_Error as e:        
            log.debug(repr(e))
            textval = repr(e)
        except Exception as e:
            log.error(repr(e))
            textval = repr(e)
        with context.push(
            field_textval=textval, 
            field_labelval=labelval, 
            field_linkval=linkval,
            field_continuation_param=linkcont):
            try:
                result = self._template.render(context)
            except Exception as e:
                log.error(repr(e))
                result = repr(e)
        log.debug("Select_view_renderer.render: result %r"%(result,))
        return result

class Select_edit_renderer(object):
    """
    Render select value for editing using supplied template
    """
    def __init__(self, template):
        self._template = Template(template)
        return

    def render(self, context):
        try:
            val     = get_field_edit_value(context, None) or ""
            # Use refer-to type if value does not include type..
            typval  = fill_type_entity_id(val, context['field']['field_ref_type'])
            textval = SelectValueMapper.encode(typval)
            options = context['field']['options']
            # options is a list of FieldChoice values
            # print repr(options)
            if textval not in [ o.value for o in options ]:
                options = list(options)                   # clone
                options.insert(0, FieldChoice(textval))   # Add missing current value to options
            with context.push(encoded_field_value=textval, field_options=options):
                result = self._template.render(context)
        except Exception as e:
            log.exception("Exception in Select_edit_renderer.render")
            log.error("Select_edit_renderer.render: "+repr(e))
            # log.error("Field val %r"%(val,))
            # log.error("Field name %r"%(context['field']['field_name'],))
            # log.error("Field type ref %r"%(context['field']['field_ref_type'],))
            # ex_type, ex, tb = sys.exc_info()
            # traceback.print_tb(tb)
            result = repr(e)
        return result

#   ----------------------------------------------------------------------------
#
#   Return render objects for select or choice controls (with or without '+' button)
#
#   ----------------------------------------------------------------------------

def get_select_renderer():
    """
    Return field renderer object for value selector (with '+' button)
    """
    return RenderFieldValue(
        view_renderer=Select_view_renderer(view_select),
        edit_renderer=Select_edit_renderer(edit_select),
        )

def get_choice_renderer():
    """
    Return field renderer object for value selector (without '+' button)
    """
    return RenderFieldValue(
        view_renderer=Select_view_renderer(view_choice),
        edit_renderer=Select_edit_renderer(edit_choice),
        )

def get_entitytype_renderer():
    """
    Return field renderer object for entitytype
    """
    return RenderFieldValue(
        view_renderer=Select_view_renderer(view_entitytype),
        edit_renderer=Select_edit_renderer(edit_entitytype),
        )

def get_view_choice_renderer():
    """
    Return field renderer object for entitytype
    """
    return RenderFieldValue(
        view_renderer=Select_view_renderer(view_view_choice),
        edit_renderer=Select_edit_renderer(edit_view_choice),
        )

# End.
